"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import logging

from loom.eka import Circuit
from loom.eka.operations import MeasureBlockSyndromes

from .generate_syndromes import generate_syndromes
from .generate_detectors import generate_detectors
from ..interpretation_step import InterpretationStep


logging.basicConfig(format="%(name)s - %(levelname)s - %(message)s")
log = logging.getLogger(__name__)


def measureblocksyndromes(
    interpretation_step: InterpretationStep,
    operation: MeasureBlockSyndromes,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Measure the syndromes of all stabilizers in a block.

    The algorithm is the following:
    - A.) Begin MeasureBlockSyndromes composite operation session

    - B.) Get stabilizers and syndrome circuit templates from the block
        - B.1) Get the stabilizers from the block
        - B.2) Get the stabilizer circuit templates from the block 
        
    - C.) Resolve circuits with actual channels (quantum and classical)
        - C.1) Find the channels for the qubits (create them if they don't exist) and \
        keep track of these in the right order
        - C.2) Get the classical bit labels from the syndrome circuit

    - D.) For each cycle:
        - D.1) Create classical channels and measurement records
        - D.2) Clone syndrome circuits and remap channels
        - D.3) Weave circuits together into a single time slice
            - NOTE: We currently assume that the circuits are constructed in order, 
              this is the responsibility of the user
        - D.4) Append woven circuit to interpretation step
        - D.5) Generate new syndromes for the stabilizers
        - D.6) Create new detectors for the syndromes
        - D.7) Update interpretation step with syndromes and detectors

    - E.) End MeasureBlockSyndromes composite operation session and append circuit

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the block from which the 
        syndrome should be measured.
    operation : MeasureBlockSyndromes
        Syndrome measurement operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        Interpretation step after the syndrome measurement operation.
    """

    # A.) Begin MeasureBlockSyndromes composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"measure {operation.input_block_name} syndromes "
            f"{operation.n_cycles} time(s)"
        ),
    )

    # B.) Get stabilizers and syndrome circuit templates from the block
    # B.1) Get the stabilizers from the block
    block = interpretation_step.get_block(operation.input_block_name)
    stabilizers = block.stabilizers

    # B.2) Get the stabilizer circuit templates from the block
    syndrome_circuit_uuids = [
        block.stabilizer_to_circuit[stabilizer.uuid] for stabilizer in stabilizers
    ]
    syndrome_circuits_templates = [
        syndrome_circuit
        for id in syndrome_circuit_uuids
        for syndrome_circuit in block.syndrome_circuits
        if syndrome_circuit.uuid == id
    ]

    # C.) Resolve circuits with actual channels (quantum and classical)
    # C.1) Find the channels for the qubits (create them if they don't exist)
    #       and keep track of these in the right order
    data_channels = [
        [
            interpretation_step.get_channel_MUT(q, "quantum")
            for q in map(str, stab.data_qubits)
        ]
        for stab in stabilizers
    ]
    ancilla_channels = [
        [
            interpretation_step.get_channel_MUT(q, "quantum")
            for q in map(str, stab.ancilla_qubits)
        ]
        for stab in stabilizers
    ]
    # C.2) Get the classical bit labels from the syndrome circuit
    cbit_labels = [
        [str(q) for q in stabilizer.ancilla_qubits] for stabilizer in stabilizers
    ]

    # D.) For each cycle
    for idx in range(operation.n_cycles):
        # D.1) Create classical channels and measurement records
        cbit_channels, measurements = [], []
        for each_cbit_label in cbit_labels:
            cbit = interpretation_step.get_new_cbit_MUT("c_" + each_cbit_label[0])
            cbit_channels.append(
                interpretation_step.get_channel_MUT(
                    f"{cbit[0]}_{str(cbit[1])}", "classical"
                )
            )
            measurements.append(cbit)
        measurements = tuple((m,) for m in measurements)

        # D.2) Clone syndrome circuits and remap channels
        mapped_syndrome_circuits = [
            syndrome_circuit.circuit.clone(
                data_channels[i] + ancilla_channels[i] + [cbit_channels[i]]
            )
            for i, syndrome_circuit in enumerate(syndrome_circuits_templates)
        ]

        # D.3) Weave circuits together into a single time slice
        if not all(
            len(each_syndrome_circuit.circuit)
            == len(mapped_syndrome_circuits[0].circuit)
            for each_syndrome_circuit in mapped_syndrome_circuits
        ):
            raise ValueError("All syndrome circuits must be of the same length.")

        woven_circuit_seq = []
        for i in range(len(mapped_syndrome_circuits[0].circuit)):
            time_slice = [
                gate
                for circuit in mapped_syndrome_circuits
                for gate in circuit.circuit[i]
            ]
            woven_circuit_seq.append(time_slice)

        # D.4) Append woven circuit to interpretation step
        interpretation_step.append_circuit_MUT(
            Circuit(
                name=f"measure {block.unique_label} syndromes - cycle {idx}",
                circuit=woven_circuit_seq,
            ),
            same_timeslice=False,
        )

        # D.5) Generate new syndromes for the stabilizers
        new_syndromes = generate_syndromes(
            interpretation_step,
            stabilizers,
            block,
            measurements,
        )

        # D.6) Create new detectors for the syndromes
        new_detectors = generate_detectors(interpretation_step, new_syndromes)

        # D.7) Update interpretation step with syndromes and detectors
        interpretation_step.append_syndromes_MUT(new_syndromes)
        interpretation_step.append_detectors_MUT(new_detectors)

    # E.) End MeasureBlockSyndromes composite operation session and append circuit
    woven_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(woven_circuit, same_timeslice)

    # For debugging purposes, unroll and log the final circuit
    unrolled_woven_circuit = Circuit(
        name=woven_circuit.name, circuit=Circuit.unroll(woven_circuit)
    )
    log.debug(unrolled_woven_circuit.detailed_str())

    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""


class EngineRunError(BaseException):
    """An Error has occurred when trying to run the Engine."""


class PropagationError(BaseException):
    """An Error has occurred when trying to propagate a PauliFrame"""


class InvalidTableauError(BaseException):
    """
    An Error has occurred when trying to validate the current Tableau during runtime.
    """


class TableauSizeError(BaseException):
    """Raised when the Tableau provided 2 tableaus do not have the same size."""


class ClassicalRegisterError(BaseException):
    """Raised when trying to perform an action with the classical register."""


class ClassicalOperationError(BaseException):
    """Raised when trying to perform a Classical Operation."""



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from .auxcnot import auxcnot
from .grow import grow
from .logical_phase_via_ywall import logical_phase_via_ywall
from .merge import merge

from .move_corners import move_corners
from .move_block import move_block
from .rsc_applicator import RotatedSurfaceCodeApplicator
from .shrink import shrink
from .split import split
from .state_injection import state_injection

# from .transversalhadamard import transversalhadamard
from .y_wall_out import y_wall_out



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from .pauli_frame import PauliFrame
from .classicalreg import ClassicalRegister


class DataStore:
    """
    An Object that records relevant information users might want recorded from
    a particular computation.
    """

    def __init__(self):
        self.measurements = {"time_step": []}
        self.pf_records = {"forward": {"time_step": []}, "backward": {"time_step": []}}
        self.cr_records = {"time_step": []}

    def __str__(self):
        return str(self.measurements)

    def set_time_step(self, time_step: int) -> None:
        """
        Sets the time step of the current Moment being processed.
        """
        self.time_step = time_step  # pylint: disable=attribute-defined-outside-init

    def record_measurements(
        self,
        measurement_id: str,
        measurement_result: int,
        is_result_random: int,
    ) -> None:
        """
        Records the measurement result in the DataStore. Raises a ValueError if
        no time step was found in the Moment.
        """
        if self.time_step is not None:
            meas_result = {
                "measurement_result": measurement_result,
                "is_random": is_result_random,
            }
            # If the measurements of a particular time step has not been
            # recorded. Record the current measurement results in that
            # time step.
            if self.time_step not in self.measurements["time_step"]:
                self.measurements["time_step"].append(self.time_step)
                self.measurements.update({str(self.time_step): {}})
            self.measurements[str(self.time_step)].update({measurement_id: meas_result})
        else:
            raise ValueError("The time step of the Moment cannot be None.")

    def record_pauli_frame(
        self,
        init_pauliframe: PauliFrame,
        recorded_pauliframe: PauliFrame,
    ) -> None:
        """
        Records the PauliFrame in the DataStore. Raises a ValueError if
        no time step was found in the Moment.
        """
        if self.time_step is not None:
            direction = init_pauliframe.direction
            pf_id = init_pauliframe.id
            record_pf = {
                "initial_pauli_frame": init_pauliframe,
                "recorded_pauli_frame": recorded_pauliframe,
            }
            if self.time_step not in self.pf_records[direction]["time_step"]:
                self.pf_records[direction]["time_step"].append(self.time_step)
                self.pf_records[direction].update({str(self.time_step): {}})
            self.pf_records[direction][str(self.time_step)].update({pf_id: record_pf})

        else:
            raise ValueError("The time step of the Moment cannot be None.")

    def record_classical_register(self, input_classical_reg: ClassicalRegister) -> None:
        """
        Records the Snapshot of the Classical Register into the DataStore.
        """
        if self.time_step is not None:
            cr_snapshot = {
                input_classical_reg.name: input_classical_reg.create_snapshot()
            }
            # Records the state of a classical register at a particular time step.
            if self.time_step not in self.cr_records["time_step"]:
                self.cr_records["time_step"].append(self.time_step)
                self.cr_records.update({str(self.time_step): {}})
            self.cr_records[str(self.time_step)].update(cr_snapshot)
        else:
            raise ValueError("The time step of the Moment cannot be None.")

    def record_measurement_from_pauliframes(
        self,
        measurement_id: str,
        results: list[int],
        init_pauliframes: list[PauliFrame],
    ) -> None:
        """
        Records measurement flips that result from the interaction of
        Pauli frames and measurement operations.

        Measurement flips is passed as a list as there could be multiple Pframes.
        """
        if self.time_step is not None:

            # first we check whether "time_step" is already written into datastore
            # if not, then we update the dict.
            if self.time_step not in self.measurements["time_step"]:
                self.measurements["time_step"].append(self.time_step)
                self.measurements.update({str(self.time_step): {}})

                self.measurements[str(self.time_step)].update(
                    {
                        measurement_id: {
                            "flip_results": {},
                        }
                    }
                )
            # if "time_step" already is present, then we open a new dict keyed by
            # "flip_results" for particular measurement_id. If "measurement_id" has not
            # been previously processed update with key.
            elif measurement_id not in self.measurements[str(self.time_step)]:
                self.measurements[str(self.time_step)].update(
                    {
                        measurement_id: {
                            "flip_results": {},
                        }
                    }
                )

            elif (
                "flip_results"
                not in self.measurements[str(self.time_step)][measurement_id]
            ):
                self.measurements[str(self.time_step)][measurement_id].update(
                    {"flip_results": {}}
                )

            for pframe, result in zip(init_pauliframes, results, strict=True):
                self.measurements[str(self.time_step)][measurement_id][
                    "flip_results"
                ].update({pframe.id: result})
        else:
            raise ValueError("The time step of the Moment cannot be None.")

    def get_pframes(
        self,
        temporal_direction: str = "forward",
    ) -> tuple[dict]:
        """
        A helper function to return a tuple of all Pframes (dict)

        `temporal_direction` can be either "forward" or "backward".

        returns:
            tuple: A tuple of dictionaries, where each dictionary
            corresponds to a PauliFrame that was created in the circuit,
            and the computed output --- for frames propagated forwards.
        """
        raw_data = self.pf_records[temporal_direction]
        raw_data_keys = raw_data["time_step"]

        return tuple(fr for key in raw_data_keys for fr in raw_data[str(key)].values())

    def get_pframe_measurements(
        self,
        pauliframe_id: str,
    ) -> dict:
        """
        Helper function to aggregate all measurement results
        belonging to the same Pframe. The keys of returned dict
        are the uuids of meausurements in the circuit.

        returns:
            dict: A dictionary where keys are measurement ids and values are
            the PauliFrame flip results for the given Pframe id.
        """

        result_data = {}
        raw_data_timekeys = self.measurements["time_step"]

        for ts in raw_data_timekeys:
            for m_id in self.measurements[str(ts)].keys():

                measurement_entry = self.measurements[str(ts)][m_id]
                if "flip_results" in measurement_entry:
                    for pf_id_sel, m_val in measurement_entry["flip_results"].items():
                        if pf_id_sel == pauliframe_id:
                            result_data.update({m_id: m_val})

        return result_data



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.interpreter.applicator.code_applicator import measureblocksyndromes
from loom.eka.operations import Grow, Shrink, MeasureBlockSyndromes
from loom.interpreter.interpretation_step import InterpretationStep
from loom.eka.utilities import Direction, Orientation, SyndromeMissingError

from .grow import grow
from .shrink import shrink
from .move_corners import move_corners
from .y_wall_out import y_wall_out
from ..operations import LogicalPhaseViaYwall
from ..code_factory import RotatedSurfaceCode


# pylint: disable=too-many-statements, too-many-locals, too-many-branches
def logical_phase_via_ywall(
    interpretation_step: InterpretationStep,
    operation: LogicalPhaseViaYwall,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Apply the logical phase via y-wall operation.
    The algorithm is the following:

    - A) Begin LogicalPhaseViaYwall composite operation session
    - B) Run consistency checks
    - C) Relocate the x logical operator to the appropriate position
    - D) Grow the block
    - E) Move a corner to prepare for the y-wall operation
    - F) Measure the syndromes
    - G) Apply the y_wall_out operation
    - H) Move all topological corners back to their geometric positions and \
    potentially grow the block towards the initial position
    - I) Shrink the block
    - J) Relocate the x logical operator to the initial position
    - K) End the composite operation session and append the circuit

    Note that whenever the logical operator is moved, there may be some extra cycles of
    measurement of the block syndromes if there are missing syndromes.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The current state of the interpretation step.
    operation : LogicalPhaseViaYwall
        The operation to be applied.
    same_timeslice : bool
        Whether to apply the operation in the same timeslice or not.
    debug_mode : bool
        Whether to run in debug mode or not.

    Returns
    -------
    InterpretationStep
        The updated interpretation step after applying the operation.
    """

    # A) Begin LogicalPhaseViaYwall composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(
            f"LogicalPhaseViaYwall on block {operation.input_block_name} "
            f"towards the {operation.growth_direction.name}."
        ),
    )

    # B) Run consistency checks
    # Check that the RotatedSurfaceCode block is square and has odd dimensions
    # Check that the topological corners coincide with the geometric corners
    # Store the initial block and the current block, initial_block is a reference that
    # will never be modified while current_block will be re-assigned during the
    # operation.
    init_block = current_block = check_consistency(interpretation_step, operation)

    growth_direction = operation.growth_direction
    init_x_log_operator = init_block.logical_x_operators[0]

    # Extract operation parameters after the checks
    distance = init_block.size[0]
    is_growth_towards_negative = growth_direction in [Direction.LEFT, Direction.TOP]
    is_init_top_left_bulk_stab_x = (
        init_block.upper_left_4body_stabilizer.pauli[0] == "X"
    )
    is_top_left_bulk_stab_x = is_growth_towards_negative ^ is_init_top_left_bulk_stab_x
    is_x_boundary_horizontal = init_block.x_boundary == Orientation.HORIZONTAL

    # C) Relocate the x logical operator to the appropriate position
    # Depending on the top-left bulk stabilizer and the x boundary orientation,
    # the top-left qubit of the x logical operator needs to be moved to a
    # different position such that after the block is grown, the x logical operator
    # is in the correct position for the y-wall operation.
    match (is_top_left_bulk_stab_x, is_x_boundary_horizontal):
        case (False, False):
            # Top-left qubit
            new_x_log_op_top_left_qubit = max(
                current_block.data_qubits, key=lambda x: -x[0] - x[1]
            )
        case (True, False):
            # Top-right qubit
            new_x_log_op_top_left_qubit = max(
                current_block.data_qubits, key=lambda x: +x[0] - x[1]
            )
        case (True, True):
            # Bottom-left qubit
            new_x_log_op_top_left_qubit = max(
                current_block.data_qubits, key=lambda x: -x[0] + x[1]
            )
        case (False, True):
            # Top-left qubit
            new_x_log_op_top_left_qubit = max(
                current_block.data_qubits, key=lambda x: -x[0] - x[1]
            )
    # Move the x logical operator to the new position
    try:
        interpretation_step, current_block = move_logical(
            interpretation_step, current_block, new_x_log_op_top_left_qubit, "X"
        )
    except SyndromeMissingError:
        # Measure block syndromes to get missing syndromes
        interpretation_step = measureblocksyndromes(
            interpretation_step,
            MeasureBlockSyndromes(init_block.unique_label, distance),
            same_timeslice=False,
            debug_mode=debug_mode,
        )
        # Move the x logical operator to the new position
        interpretation_step, current_block = move_logical(
            interpretation_step, current_block, new_x_log_op_top_left_qubit, "X"
        )

    # D) Grow the block
    # The block is grown to the right if the x boundary is horizontal and down if
    # the x boundary is vertical
    growth_length = distance

    interpretation_step = grow(
        interpretation_step,
        Grow(init_block.unique_label, growth_direction, growth_length),
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(init_block.unique_label)

    # Measure block syndromes to finalize Grow
    interpretation_step = measureblocksyndromes(
        interpretation_step,
        MeasureBlockSyndromes(init_block.unique_label, distance),
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # If the growth direction is towards the negative direction, the Z logical operator
    # needs to be moved to the new top-left qubit of the block.
    if is_growth_towards_negative:
        # Move the Z logical operator to the new position
        interpretation_step, current_block = move_logical(
            interpretation_step,
            current_block,
            current_block.upper_left_qubit,
            pauli="Z",
        )

    # E) Move the corner to prepare for the y-wall operation
    # Depending on the top-left bulk stabilizer and the x boundary orientation,
    # a different corner qubit needs to be moved to prepare for the y-wall operation.
    current_block: RotatedSurfaceCode = interpretation_step.get_block(
        init_block.unique_label
    )
    match (is_top_left_bulk_stab_x, is_x_boundary_horizontal):
        case (False, False):
            corner_0_location_directions = [Direction.BOTTOM, Direction.LEFT]
            corner_0_move_direction = Direction.TOP
        case (True, False):
            corner_0_location_directions = [Direction.BOTTOM, Direction.RIGHT]
            corner_0_move_direction = Direction.TOP
        case (True, True):
            corner_0_location_directions = [Direction.BOTTOM, Direction.RIGHT]
            corner_0_move_direction = Direction.LEFT
        case (False, True):
            corner_0_location_directions = [Direction.TOP, Direction.RIGHT]
            corner_0_move_direction = Direction.LEFT
    corner_0_location = (
        set(current_block.boundary_qubits(corner_0_location_directions[0]))
        .intersection(current_block.boundary_qubits(corner_0_location_directions[1]))
        .pop()
    )
    # Find the direction to move the corner qubit
    corner_0_how_far = distance - 1
    # Move the corner
    corner_args = ((corner_0_location, corner_0_move_direction, corner_0_how_far),)
    interpretation_step = move_corners(
        interpretation_step,
        current_block,
        corner_args,
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # F) Measure the syndromes
    # Measure the syndromes of the block such that the block is projected onto
    # the new stabilizers before the y-wall operation is applied.
    # Measure for distance - 1 since the move_corners operation has already measured
    # the stabilizers once.
    interpretation_step = measureblocksyndromes(
        interpretation_step,
        MeasureBlockSyndromes(init_block.unique_label, distance - 1),
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # G) Apply the y_wall_out operation
    # The y_wall_out operation is applied to implement the main part of the
    # logical phase via y-wall operation.
    wall_position = distance
    wall_orientation = init_block.x_boundary.perpendicular()
    current_block = interpretation_step.get_block(init_block.unique_label)

    interpretation_step = y_wall_out(
        interpretation_step,
        current_block,
        wall_position,
        wall_orientation,
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # H) Move all topological corners back to their geometric positions and
    ## potentially grow the block towards the initial position
    # Move the topological corners back to their geometric positions
    current_block = interpretation_step.get_block(init_block.unique_label)

    # This is a geometric corner but at the incorrect position.
    # Depending on the top-left bulk stabilizer and the x boundary orientation,
    # a different corner qubit needs to be moved to the appropriate direction.
    match (is_top_left_bulk_stab_x, is_x_boundary_horizontal):
        case (False, False):
            corner_1_location_directions = [Direction.BOTTOM, Direction.RIGHT]
            corner_1_move_direction = Direction.LEFT
        case (True, False):
            corner_1_location_directions = [Direction.BOTTOM, Direction.LEFT]
            corner_1_move_direction = Direction.RIGHT
        case (True, True):
            corner_1_location_directions = [Direction.TOP, Direction.RIGHT]
            corner_1_move_direction = Direction.BOTTOM
        case (False, True):
            corner_1_location_directions = [Direction.BOTTOM, Direction.RIGHT]
            corner_1_move_direction = Direction.TOP
    corner_1_location = (
        set(current_block.boundary_qubits(corner_1_location_directions[0]))
        .intersection(current_block.boundary_qubits(corner_1_location_directions[1]))
        .pop()
    )
    corner_1_move_how_far = distance - 1
    # Corner 1 arguments
    corner_1_args = (corner_1_location, corner_1_move_direction, corner_1_move_how_far)

    # Move corner 2
    # This is a topological corner but at the incorrect position.
    current_block = interpretation_step.get_block(init_block.unique_label)
    # Find the topological corner that is not located at the geometric corner
    corner_2_location = (
        set(current_block.topological_corners) - set(current_block.geometric_corners)
    ).pop()
    # It needs to be moved in the opposite direction of corner 0
    corner_2_move_direction = corner_0_move_direction.opposite()
    corner_2_how_far = distance - 1
    # Corner 2 arguments
    corner_2_args = (corner_2_location, corner_2_move_direction, corner_2_how_far)

    # Move the corners one-by-one
    for corner_args in (corner_1_args, corner_2_args):
        # Move the corner
        interpretation_step = move_corners(
            interpretation_step=interpretation_step,
            block=current_block,
            corner_args=(corner_args,),
            same_timeslice=False,
            debug_mode=debug_mode,
        )
        # Obtain the new current_block
        current_block = interpretation_step.get_block(init_block.unique_label)
        # Measure block syndromes to ensure that the moving of corners is FT
        # move_corners already measured the syndromes once, so we can just measure them
        # (distance - 1) times
        interpretation_step = measureblocksyndromes(
            interpretation_step,
            MeasureBlockSyndromes(init_block.unique_label, distance - 1),
            same_timeslice=False,
            debug_mode=debug_mode,
        )

    # If the growth direction is towards the negative direction, the block needs to
    # be grown by one unit towards the opposite direction of the growth direction
    # for the final shrink to leave it at the correct position.
    # NOTE: This is because y_wall_out has been implemented such that the block is
    # shrunk from the right or from the bottom.
    if is_growth_towards_negative:
        # Grow by one towards the opposite direction of the growth direction
        interpretation_step = grow(
            interpretation_step,
            Grow(init_block.unique_label, growth_direction.opposite(), 1),
            same_timeslice=False,
            debug_mode=debug_mode,
        )
        # Measure the syndromes of the block such that the block is projected onto
        # the new stabilizers after the growth.
        interpretation_step = measureblocksyndromes(
            interpretation_step,
            MeasureBlockSyndromes(init_block.unique_label, distance),
            same_timeslice=False,
            debug_mode=debug_mode,
        )

    # I) Shrink the block
    # The block is shrunk to the original size. Since the y_wall_out operation
    # shrunk the block by 1, the block needs to be shrunk by 1 unit less than the
    # growth operation.
    shrink_direction = growth_direction
    shrink_length = growth_length - (1 if not is_growth_towards_negative else 0)
    # Shrink the block
    interpretation_step = shrink(
        interpretation_step,
        Shrink(init_block.unique_label, shrink_direction, shrink_length),
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    # Measure the syndromes of the block such that the block is projected onto
    # the new stabilizers after the shrink.
    interpretation_step = measureblocksyndromes(
        interpretation_step,
        MeasureBlockSyndromes(init_block.unique_label, distance),
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # J) Relocate the x logical operator to the initial position
    # The x logical operator is moved back to its original position.
    current_block = interpretation_step.get_block(init_block.unique_label)
    # Find the top-left corner of the initial x logical operator
    final_x_log_op_top_left_qubit = max(
        init_x_log_operator.data_qubits, key=lambda x: -x[0] - x[1]
    )

    # Move the x logical operator back to the original position
    # Syndromes will always be available, so no need for catching exception here
    interpretation_step, current_block = move_logical(
        interpretation_step, current_block, final_x_log_op_top_left_qubit, "X"
    )

    # K) End the operation session and append the circuit
    logical_phase_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(logical_phase_circuit, same_timeslice)

    return interpretation_step


def check_consistency(
    interpretation_step: InterpretationStep, operation: LogicalPhaseViaYwall
) -> RotatedSurfaceCode:
    """
    Check that the block is consistent with the logical phase via y-wall operation.

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The InterpretationStep that contains the block to which the operation is
        applied.
    operation: LogicalPhaseViaYwall
        The operation to be applied.

    Returns
    -------
    RotatedSurfaceCode
        The block to which the operation is applied.

    Raises
    ------
    ValueError
        If the block is not square and has does not have odd dimensions.
        If the topological corners do not coincide with the geometric corners.
    """
    # Obtain the info from interpretation step
    block = interpretation_step.get_block(operation.input_block_name)
    if not isinstance(block, RotatedSurfaceCode):
        raise ValueError(
            f"The LogicalPhaseViaYwall operation can only be applied to "
            f"RotatedSurfaceCode blocks. The block {block.unique_label} "
            f"is a {block.__class__.__name__} block."
        )

    dim_x, dim_y = block.size
    if dim_x != dim_y or dim_x % 2 != 1:
        raise ValueError(
            "Block must be square and have odd dimensions for the logical "
            "phase via y-wall operation."
        )

    if set(block.topological_corners) != set(block.geometric_corners):
        raise ValueError(
            "Topological corners must coincide with geometric corners for "
            "the logical phase via y-wall operation."
        )

    growth_direction = operation.growth_direction
    if growth_direction.to_orientation() != block.x_boundary:
        raise ValueError(
            f"The growth direction ({growth_direction.name}) must be parallel to the x "
            f"boundary ({block.x_boundary.name}) of the block for the logical phase "
            "via y-wall operation."
        )

    return block


def move_logical(
    interpretation_step: InterpretationStep,
    current_block: RotatedSurfaceCode,
    new_up_left_qubit: tuple[int, int, int],
    pauli: str,
) -> tuple[InterpretationStep, RotatedSurfaceCode]:
    """
    Move the logical operator to a new position. If the logical operator is
    already in the new position, no action is taken. If the logical operator
    is not in the new position, it is moved to the new position and the stabilizers are
    redefined. We assume that the block is square and its logical operator is a
    straight line. Can be used for both X and Z logical operators.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The current state of the interpretation step.
    current_block : RotatedSurfaceCode
        The block to which the operation is applied.
    new_up_left_qubit : tuple[int, int, int]
        The new position of the top-left qubit of the logical x operator.
    pauli : str
        The logical operator to be moved. Can be either "X" or "Z".

    Returns
    -------
    interpretation_step : InterpretationStep
        The updated interpretation step after applying the operation.
    current_block : RotatedSurfaceCode
        The updated block after applying the operation.

    Raises
    ------
    SyndromeMissingError
        If syndromes required to redefine the logical x operator are missing.
    """
    if pauli == "X":
        current_logical_op = current_block.logical_x_operators[0]
    elif pauli == "Z":
        current_logical_op = current_block.logical_z_operators[0]
    else:
        raise ValueError(
            f"Logical operator {pauli} is not supported. Only 'X' and 'Z' are "
            "supported."
        )

    # Use method to obtain the new logical operator and the stabilizers
    # that need to be redefined
    new_logical, stabilizers_to_redefine = (
        current_block.get_shifted_equivalent_logical_operator(
            current_logical_op, new_up_left_qubit
        )
    )

    # If the logical x operator is not in the new position, we need to redefine
    # the logical x operator and declare its evolution
    if stabilizers_to_redefine:
        # Check that the stabilizers to redefine have syndromes associated
        # with the input block. This is to ensure that the logical phase
        # NOTE: At this point a `SyndromeMissingError` is raised if the syndromes are
        # not associated with the input block.
        redefinition_stab_cbits = interpretation_step.retrieve_cbits_from_stabilizers(
            stabilizers_to_redefine, current_block
        )

        # Find the evolution of the logical operators
        if pauli == "X":
            interpretation_step.logical_x_evolution[new_logical.uuid] = (
                current_block.logical_x_operators[0].uuid,
            ) + tuple(stab.uuid for stab in stabilizers_to_redefine)

            # Define the new logical x/z operator
            # (only the logical x operator is moved)
            new_logical_x = new_logical
            new_logical_z = current_block.logical_z_operators[0]
        else:
            interpretation_step.logical_z_evolution[new_logical.uuid] = (
                current_block.logical_z_operators[0].uuid,
            ) + tuple(stab.uuid for stab in stabilizers_to_redefine)

            # Define the new logical z/x operator
            # (only the logical z operator is moved)
            new_logical_x = current_block.logical_x_operators[0]
            new_logical_z = new_logical

        # Pass the logical updates onto the new logical operator
        # along with the Cbits from the stabilizers that are need to redefine it
        interpretation_step.update_logical_operator_updates_MUT(
            pauli, new_logical.uuid, redefinition_stab_cbits, inherit_updates=True
        )

        new_block = RotatedSurfaceCode(
            stabilizers=current_block.stabilizers,
            logical_x_operators=(new_logical_x,),
            logical_z_operators=(new_logical_z,),
            syndrome_circuits=current_block.syndrome_circuits,
            stabilizer_to_circuit=current_block.stabilizer_to_circuit,
            unique_label=current_block.unique_label,
        )
        interpretation_step.update_block_history_and_evolution_MUT(
            (new_block,), (current_block,)
        )
        current_block = new_block

    return interpretation_step, current_block



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=too-many-lines

from __future__ import annotations
from itertools import product
import numpy as np
import networkx as nx


from . import matrices  # pylint: disable=cyclic-import
from .stabilizer import Stabilizer


def cartesian_product_tanner_graphs(
    t_graph_1: ClassicalTannerGraph, t_graph_2: ClassicalTannerGraph
) -> TannerGraph:
    """
    Computes the Cartesian product of two ClassicalTannerGraphs Gi(Vi,Ci,Ei), where Vi
    is the set of data nodes, Ci is the set of check nodes and Ei is the set of edges.
    Each ClassicalTannerGraph is associated with a classical code. See arXiv:2109.14609
    for a detailed description of the construction.

    The cartesian product yields a TannerGraph G(V,C,E) describing a quantum CSS code.
    The nodes of the product TannerGraph are labelled via two coordinates (v1,v2), each
    inherited from each classical code. The set of data nodes V is built out of the
    union of the nodes resulting from the product of the data nodes (v1,v2)  and
    of the check nodes (c1,c2) of each classical code. The set of check nodes C follows
    from the product of check and data nodes from each classical codes, i.e. (c1,v2) and
    (v1,c2). We assign X check nodes to be the first partition of check nodes and Z
    checks to be the second one. Edges are drawn between nodes (a1,a2) and (b1,b2), if
    either a2=b2 and G1 contains the edge (a1,b1) or a1=b1 and G2 contains the edge
    (a2,b2). The edges of product Tanner graphs are decorated with a cardinality label
    {`N`,`S`,`E`,`W`} according to the definitions introduced in arXiv:2109.14609.

    Note that we assign the first and second input ClassicalTannerGraphs to X and Z
    checks respectively, regardless of their associated `check_type` attribute.
    Furthermore, we impose a single integer labelling of the nodes, of the form `(i,)`
    in the input ClassicalTannerGraphs. If the any of the inputs does not satisfy this
    requirement, we build a relabelled graph using the `relabelled_graph()` method of
    the ClassicalTannerGraph class. While this constraint is fundamental from the
    cartesian product perspective, it is required in order to decorate the edges with
    cardinal labels.

    Parameters
    ----------
    t_graph_1: ClassicalTannerGraph
        Tanner graph of the first classical code.
    t_graph_2: ClassicalTannerGraph
        Tanner graph of the second classical code.

    Returns
    -------
    tanner_product: TannerGraph
        Tanner graph from cartesian product of t_graph_1 and t_graph_2, describing the
        full HGP code.
    """

    # Check inputs are of valid type
    if not all(
        isinstance(t_graph, ClassicalTannerGraph) for t_graph in [t_graph_1, t_graph_2]
    ):
        raise TypeError("Both inputs must be of type ClassicalTannerGraph.")

    # Ensure single integer tuple labelling of nodes
    # If not satisfied, generate relabelled graph
    if any(
        not isinstance(n, tuple) or (len(n) != 1) or (not isinstance(n[0], int))
        for n in t_graph_1.graph.nodes()
    ):
        t_graph_1 = t_graph_1.relabelled_graph()

    if any(
        not isinstance(n, tuple) or (len(n) != 1) or (not isinstance(n[0], int))
        for n in t_graph_2.graph.nodes()
    ):
        t_graph_2 = t_graph_2.relabelled_graph()

    # Initialize cartesian product graph
    graph_product = nx.Graph()

    # Extract size of node sets for both graphs
    len1 = len(t_graph_1.graph.nodes)
    len2 = len(t_graph_2.graph.nodes)

    # Add nodes to the product graph
    for t1_node, t2_node in product(
        t_graph_1.graph.nodes(data=True), t_graph_2.graph.nodes(data=True)
    ):
        # Nodes are indexed with a tuple, inherited from the indices in
        # the original codes
        node_pair = (t1_node[0][0], t2_node[0][0])
        label1 = t1_node[1]["label"]
        label2 = t2_node[1]["label"]

        # If both nodes are of the same type they are a data qubit
        if (label1 != "data" and label2 != "data") or (
            label1 == "data" and label2 == "data"
        ):
            graph_product.add_node(node_pair, label="data")

        # If nodes are different they correspond to a check node
        else:
            # We choose that if the nodes of the first code are checks, these
            # become X checks
            if label2 == "data":
                graph_product.add_node(node_pair, label="X")
            # If the nodes of the second code are checks, these become Z checks
            else:
                graph_product.add_node(node_pair, label="Z")

    # Add horizontal edges to the product graph, via edges inherited from the first code
    for (t1_node1, t1_node2), t2_node in product(
        t_graph_1.graph.edges(), t_graph_2.graph.nodes(data=False)
    ):
        # Define edges such that they follow the (check,data) order
        unordered_edge = ((t1_node1[0], t2_node[0]), (t1_node2[0], t2_node[0]))
        edge = (
            unordered_edge
            if graph_product.nodes[unordered_edge[1]]["label"] == "data"
            else unordered_edge[::-1]
        )

        # Compute their "label-distance"
        diff = edge[1][0] - edge[0][0]

        # Associate with E or W cardinality according to their label-distance
        is_east = diff % len1 <= len1 / 2
        cardinality = "E" if is_east else "W"

        # Add edge to the product
        graph_product.add_edge(*edge, cardinality=cardinality)

    # Add vertical edges to the product graph, via edges inherited from the second code
    for t1_node, (t2_node1, t2_node2) in product(
        t_graph_1.graph.nodes(data=False), t_graph_2.graph.edges()
    ):
        # Define edges such that they follow the (check,data) order
        unordered_edge = ((t1_node[0], t2_node1[0]), (t1_node[0], t2_node2[0]))
        edge = (
            unordered_edge
            if graph_product.nodes[unordered_edge[1]]["label"] == "data"
            else unordered_edge[::-1]
        )

        # Compute their "label-distance"
        diff = edge[1][1] - edge[0][1]

        # Associate with N or S cardinality according to their label-distance
        is_north = diff % len2 <= len2 / 2
        cardinality = "N" if is_north else "S"

        # Add edge to the product
        graph_product.add_edge(*edge, cardinality=cardinality)

    # Convert into TannerGraph object
    tanner_product = TannerGraph(graph_product)

    return tanner_product


def verify_css_code_stabilizers(
    stabilizers: list[Stabilizer] | tuple[Stabilizer],
) -> bool:
    """
    Verifies that the list of stabilizers define a valid CSS code.

    Parameters
    ----------
    stabilizers : list[Stabilizer] | tuple[Stabilizer]
        List of stabilizers defining a quantum CSS code.

    Returns
    -------
    valid: bool
        If True, the stabilizers define a valid CSS code.
    """

    # Check input
    if not isinstance(stabilizers, list) and not isinstance(stabilizers, tuple):
        raise TypeError("Input must be a list or tuple.")

    if len(stabilizers) == 0:
        raise ValueError("No stabilizers provided.")

    if not all(isinstance(stabilizer, Stabilizer) for stabilizer in stabilizers):
        raise TypeError("Input must be a list or tuple of Stabilizer objects.")

    # Verify all stabilizers are either of X or Z type
    valid = all(
        set(stabilizer.pauli) == {"X"} or set(stabilizer.pauli) == {"Z"}
        for stabilizer in stabilizers
    )

    # If all stabilizers are X and Z, verify that they all commute
    if valid:
        x_stabs = [stab for stab in stabilizers if set(stab.pauli) == {"X"}]
        z_stabs = [stab for stab in stabilizers if set(stab.pauli) == {"Z"}]
        for x_stab in x_stabs:
            for z_stab in z_stabs:
                if not x_stab.commutes_with(z_stab):
                    raise ValueError(
                        f"Input Stabilizers {x_stab} and {z_stab} do not commute."
                    )

    return valid


class ClassicalTannerGraph:
    """Classical Tanner Graph representation."""

    def __init__(
        self,
        input: nx.Graph | tuple[Stabilizer, ...] | matrices.ClassicalParityCheckMatrix,
        # pylint: disable=redefined-builtin
    ):
        """The Classical Tanner Graph class stores a bipartite graph representation of a
        classical error-correcting code. The graph is bipartite, with data nodes on one
        partition and check nodes on the other. The check nodes can be labelled with
        either 'X' or 'Z' if the code is to be related with quantum error properties,
        or it can be generically labelled as "check". If a networkx.Graph is given as
        input, the graph will go through a set of verifications to ensure its validity
        to represent the classical code. The input can also be a tuple of stabilizers,
        which need to be of the same time pauli type, and the graph will be generated
        from them. The graph can be converted into a list of stabilizers. The check type
        of the Tanner graph can be modified using the set_check_type method. Lastly, the
        class can also be instantiated from ClassicalParityCheckMatrix object, where the
        matrix is mapped to a Tanner graph and the check nodes are labeled as 'check'.

        Parameters
        ----------
        input : nx.Graph | tuple[Stabilizer,...] | ClassicalParityCheckMatrix
            Input graph, tuple of stabilizers, or ClassicalParityCheckMatrix to
            build the Classical Tanner graph.
        """
        self.check_type = None
        # Initialize graph based on input type
        if isinstance(input, nx.Graph):
            self.verify_input_graph(input)
            self.graph = input
        elif isinstance(input, tuple) and all(
            isinstance(item, Stabilizer) for item in input
        ):
            self.graph = self.generate_graph_from_stabilizers(input)
        elif isinstance(input, matrices.ClassicalParityCheckMatrix):
            self.graph = self.generate_graph_from_matrix(input)
        else:
            raise TypeError(
                "A networkx.Graph, a tuple of Stabilizer or a "
                "ClassicalParityCheckMatrix must be provided."
            )

        # Assign data nodes
        self.data_nodes = [
            n for n, attr in self.graph.nodes(data=True) if attr["label"] == "data"
        ]

        # Assign check nodes
        self.check_nodes = [
            n
            for n, attr in self.graph.nodes(data=True)
            if attr["label"] == self.check_type
        ]

    def verify_input_graph(self, graph: nx.Graph) -> None:
        """Verify the input graph is a faithful representation of a Classical
        Tanner graph.

        Parameters
        ----------
        graph : nx.Graph
            Input graph to be verified.
        """

        # Check graph is not empty
        if len(graph) == 0:
            raise ValueError("Input graph is empty. Please provide a non-empty graph.")

        # Extract labelling of nodes
        node_attributes = [attr for _, attr in graph.nodes(data=True)]

        # Check for unlabelled nodes
        if not all("label" in attr for attr in node_attributes):
            raise ValueError(
                "Missing node labels. All nodes should contain a 'label' "
                "attribute, with values 'data', 'X', 'Z' or 'check'."
            )

        # Check for invalid labels
        labels = {attr["label"] for attr in node_attributes}

        # There should be only two labels "data" and "X","Z" or "check" for check nodes
        if labels not in [{"data", "X"}, {"data", "Z"}, {"data", "check"}]:
            raise ValueError(
                "Invalid node labels in the input graph. Must be 'data' for data nodes"
                " and only 'X', 'Z' or 'check' for check nodes."
            )

        # Set check type based on left over label
        self.check_type = next(iter(labels - {"data"}))

        # Verify input graph is bipartite properties by checking connected components
        for component in nx.connected_components(graph):
            # Extract subcomponent
            sub_g = graph.subgraph(component)

            if not nx.algorithms.bipartite.is_bipartite(sub_g):
                raise ValueError("Graph is not bipartite.")

            # Verify partitions are correctly labelled
            part1, part2 = nx.algorithms.bipartite.sets(sub_g)

            # Ensure each partition contains only one type of node
            part1_labels = {graph.nodes[node]["label"] for node in part1}
            part2_labels = {graph.nodes[node]["label"] for node in part2}

            # Raise error if one partition contains more than one type of node
            if (len(part1_labels) > 1 and len(part2_labels) == 1) or (
                len(part2_labels) > 1 and len(part1_labels) == 1
            ):
                inter = next(iter(part1_labels & part2_labels))
                raise ValueError(
                    f"Graph contains invalid edges between" f" '{inter}' nodes."
                )
            # Raise error if both partitions contain more than one type of node
            if len(part1_labels) > 1 and len(part2_labels) > 1:
                raise ValueError(
                    f"Graph contains invalid edges between 'data' nodes and"
                    f" '{self.check_type}' nodes."
                )

    def generate_graph_from_stabilizers(
        self, stabilizers: tuple[Stabilizer]
    ) -> nx.Graph:
        """Generate a Classical Tanner graph from a tuple of stabilizers.

        Parameters
        ----------
        stabilizers : tuple[Stabilizer]
            Tuple of stabilizers to generate the Classical Tanner graph. All must be of
            the same pauli type.

        Returns
        -------
        t_graph : nx.Graph
            Classical Tanner graph, as a networkx object generated from the input
            stabilizers.
        """

        # Extract Pauli type from Stabilizers
        pauli_type = set(p for stab in stabilizers for p in stab.pauli)

        # Verify the pauli type is unique
        if len(pauli_type) > 1:
            raise ValueError(
                "Input stabilizers must be of the same type to define a classical"
                " Tanner graph."
            )

        # Assign check type based on the unique pauli type
        self.check_type = next(iter(pauli_type))

        # Verify that there is only a single ancilla per Stabilizer
        if any(len(stab.ancilla_qubits) != 1 for stab in stabilizers):
            raise ValueError("All Stabilizers must contain a single ancilla qubit.")

        # Verify that all ancillas are different
        check_qubits = {q for stab in stabilizers for q in stab.ancilla_qubits}
        if len(check_qubits) != len(stabilizers):
            raise ValueError("All ancilla qubits must be different.")

        # Extract data_qubits
        data_qubits = {q for stab in stabilizers for q in stab.data_qubits}

        # Initialize graph and add nodes
        t_graph = nx.Graph()
        t_graph.add_nodes_from(data_qubits, label="data")
        t_graph.add_nodes_from(check_qubits, label=self.check_type)

        # Generate edges
        edges = [
            (q, stab.ancilla_qubits[0])
            for stab in stabilizers
            for q in stab.data_qubits
        ]
        t_graph.add_edges_from(edges)

        return t_graph

    def to_stabilizers(self, pauli_type: str = None) -> list[Stabilizer]:
        """Convert the Classical Tanner graph to a list of stabilizers. Input pauli_type
        is required if check nodes have not been assigned a check_type. If check nodes
        have been assigned a check_type, the input pauli_type will override the
        assignment.

        Parameters
        ----------
        pauli_type : str | None
            Pauli type of the stabilizers, either 'X' or 'Z'. If None, the check_type
            of the Tanner graph will be used. Default is None.

        Returns
        -------
        stabilizers : list[Stabilizer]
            List of stabilizers generated from the Classical Tanner graph.
        """

        # Raise error if pauli_type input is missing and nodes do not have a pauli label
        if pauli_type is None and self.check_type == "check":
            raise ValueError(
                "Check nodes have not been assigned a pauli type. Please provide a"
                " pauli_type input."
            )
        # If no pauli_type is given, use the check_type
        if pauli_type is None:
            pauli_type = self.check_type
        # Ensure input pauli_type is valid
        elif pauli_type is not None and pauli_type not in ["X", "Z"]:
            raise ValueError("Pauli type must be either 'X' or 'Z'.")

        # Verify that all nodes can be converted into coordinates
        all_nodes = list(self.graph.nodes)
        if not all(
            isinstance(t, tuple)
            and len(t) == len(all_nodes[0])
            and all(isinstance(e, int) for e in t)
            for t in all_nodes
        ):
            raise ValueError(
                "Nodes are not tuples of equal size and cannot be "
                "converted to list of stabilizers. Consider using "
                "relabelled_graph() method, to generate a re-indexed graph and later"
                " convert to Stabilizer list."
            )

        # Generate stabilizers by checking the neighbors of each check node
        def generate_stabilizer(check_node):
            data_nodes = list(self.graph.neighbors(check_node))
            return Stabilizer(
                pauli=pauli_type * len(data_nodes),
                data_qubits=[data_node + (0,) for data_node in data_nodes],
                ancilla_qubits=[check_node + (1,)],
            )

        stabilizers = [generate_stabilizer(node) for node in self.check_nodes]

        return stabilizers

    def relabelled_graph(self) -> ClassicalTannerGraph:
        """Relabel a Classical Tanner graph to identify nodes with tuples of
        integers which can later be converted into Stabilizers. The nodes are a assigned
        a single-element tuple with an integer index. The data nodes are counted
        first, followed by the check nodes. The old node labels are stored in the node
        attributes under the key 'original_node'.

        Returns
        -------
        relabelled_tanner : ClassicalTannerGraph
            Relabelled Classical Tanner graph.
        """

        # Initialize graph
        t_graph = nx.Graph()

        # Add nodes
        data_nodes_info = {
            (i,): {"label": "data", "original_node": n}
            for i, n in enumerate(self.data_nodes)
        }
        data_nodes_dict = {v["original_node"]: k for k, v in data_nodes_info.items()}

        check_nodes_info = {
            (i + len(data_nodes_info),): {"label": self.check_type, "original_node": n}
            for i, n in enumerate(self.check_nodes)
        }
        check_nodes_dict = {v["original_node"]: k for k, v in check_nodes_info.items()}

        t_graph.add_nodes_from(data_nodes_info.items())
        t_graph.add_nodes_from(check_nodes_info.items())

        # Add edges
        edges = [
            (
                (data_nodes_dict[u], check_nodes_dict[v])
                if self.graph.nodes[u]["label"] == "data"
                else (data_nodes_dict[v], check_nodes_dict[u])
            )
            for u, v in self.graph.edges()
        ]
        t_graph.add_edges_from(edges)

        # Construct new relabelled Classical Tanner graph
        relabelled_tanner = ClassicalTannerGraph(t_graph)

        return relabelled_tanner

    def set_check_type(self, new_check_type: str) -> None:
        """Modify the check type of the Tanner graph.

        Parameters
        ----------
        new_check_type : str
            New check type to be assigned to the Tanner graph. Must be either 'X', 'Z'
            or 'check'.
        """

        # Ensure correct input
        if new_check_type not in ["X", "Z", "check"]:
            raise ValueError("Check type must be either 'X', 'Z' or 'check'.")

        # Create a dictionary of updated labels
        updated_labels = {
            n: new_check_type
            for n, attr in self.graph.nodes(data=True)
            if attr["label"] == self.check_type
        }

        # Apply the updates to the graph
        nx.set_node_attributes(self.graph, updated_labels, "label")

        # Update the check type
        self.check_type = new_check_type

    def generate_graph_from_matrix(
        self, h_matrix: matrices.ClassicalParityCheckMatrix
    ) -> nx.Graph:
        """Generate Tanner graph from a ClassicalParityCheckMatrix object.

        Parameters
        ----------
        h_matrix : matrices.ClassicalParityCheckMatrix
            Parity-check matrix to generate the Tanner graph.

        Returns
        -------
        g : nx.Graph
            Classical Tanner graph, as a networkx object generated from the input
            parity-check matrix.
        """

        # Initialize graph
        g = nx.Graph()

        # Extract the dimensions of the parity-check matrix
        n_rows, n_cols = h_matrix.matrix.shape

        # Add data qubit nodes from number of columns
        data_nodes = [(i,) for i in range(n_cols)]
        g.add_nodes_from(data_nodes, label="data")

        # Add check nodes from number of rows and set check_type
        check_nodes = [(i,) for i in range(n_cols, n_cols + n_rows)]
        g.add_nodes_from(check_nodes, label="check")
        self.check_type = "check"

        # Extract the edges from the indices associated with non-vanishing terms
        for i, row in enumerate(h_matrix.matrix):
            for j in np.where(row == 1)[0]:
                g.add_edge(check_nodes[i], data_nodes[j])

        return g

    def __eq__(self, other: ClassicalTannerGraph) -> bool:
        """Check if two Classical Tanner graphs are equal. Two graphs are equal if they
        have the same nodes and edges, and the same check type.

        Parameters
        ----------
        other : ClassicalTannerGraph
            Another Classical Tanner graph to compare with.

        Returns
        -------
        bool
            True if the Classical Tanner graphs are equal, False otherwise.
        """

        if not isinstance(other, ClassicalTannerGraph):
            raise TypeError(
                "Comparison is only supported between ClassicalTannerGraph objects."
            )

        def are_node_attributes_not_the_same():
            return set(self.data_nodes) != set(other.data_nodes) or set(
                self.check_nodes
            ) != set(other.check_nodes)

        def are_check_types_not_the_same():
            return self.check_type != other.check_type

        def are_nodes_not_the_same():
            return self.graph.nodes(data=True) != other.graph.nodes(data=True)

        # Check edges by sorting the tuples to ensure order does not matter
        def are_edges_not_the_same():
            return set(tuple(sorted(edge)) for edge in self.graph.edges()) != set(
                tuple(sorted(edge)) for edge in other.graph.edges()
            )

        # Check class attributes and graph properties
        return not (
            are_node_attributes_not_the_same()
            or are_check_types_not_the_same()
            or are_nodes_not_the_same()
            or are_edges_not_the_same()
        )


class TannerGraph:
    """A class representing a Tanner graph."""

    def __init__(
        self,
        input: nx.Graph | tuple[Stabilizer] | matrices.ParityCheckMatrix,
        # pylint: disable=redefined-builtin
    ):
        """
        The TannerGraph class stores a bipartite graph representation of a quantum
        CSS code. The graph is bipartite, with data nodes on one partition and check
        nodes on the other. The check nodes can be labelled with either 'X' or 'Z',
        associated with the type of stabilizers they represent and type of errors they
        check for, e.g. Z check nodes check for X errors and vice versa. If a
        networkx.Graph is given as input, the graph will go through a set of
        verifications to ensure its validity to represent the code. The input can also
        be a tuple of stabilizers, which need to satisfy the CSS condition, i.e.
        all must be either X- or Z-type, and the graph will be generated from them.

        Parameters
        ----------
        input : nx.Graph | tuple[Stabilizer] | ParityCheckMatrix
            Input graph or tuple of stabilizers to build the Tanner graph.
        """

        if isinstance(input, nx.Graph):
            self.verify_input_graph(input)
            self.graph = input

        elif isinstance(input, tuple) and all(
            isinstance(item, Stabilizer) for item in input
        ):
            self.graph = self.generate_graph_from_stabilizers(input)

        elif isinstance(input, matrices.ParityCheckMatrix):
            self.graph = self.generate_graph_from_matrix(input)

        else:
            raise TypeError(
                "A networkx.Graph, a tuple of Stabilizers or a ParityCheckMatrix must "
                "be provided."
            )

        # Assign data nodes
        self.data_nodes = [
            n for n, attr in self.graph.nodes(data=True) if attr["label"] == "data"
        ]

        # Assign X check nodes
        self.x_nodes = [
            n for n, attr in self.graph.nodes(data=True) if attr["label"] == "X"
        ]

        # Assign Z check nodes
        self.z_nodes = [
            n for n, attr in self.graph.nodes(data=True) if attr["label"] == "Z"
        ]

    def verify_input_graph(self, graph: nx.Graph) -> None:
        """
        Verify the input graph is bipartite and that nodes are correctly labelled.

        Parameters
        ----------
        graph : nx.Graph
            Input graph to be verified.
        """

        # Check graph is not empty
        if len(graph) == 0:
            raise ValueError("Input graph is empty. Please provide a non-empty graph.")

        # Verify labelling of nodes
        node_attributes = [attr for _, attr in graph.nodes(data=True)]

        # Check for unlabelled nodes
        if not all("label" in attr for attr in node_attributes):
            raise ValueError(
                "Missing node labels. All nodes should contain a 'label' "
                "attribute, with values 'X', 'Z' or 'data'."
            )

        # Check for invalid labels
        labels = {attr["label"] for attr in node_attributes}
        if not labels.issubset({"X", "Z", "data"}):
            raise ValueError(
                "Invalid node labels in the input graph. Must be 'X', 'Z', or 'data'."
            )

        # Verify input graph is bipartite properties by checking connected components
        for component in nx.connected_components(graph):

            # Extract subcomponent
            sub_g = graph.subgraph(component)
            if not nx.algorithms.bipartite.is_bipartite(sub_g):
                raise ValueError("Graph is not bipartite.")

            # Verify partitions are correctly labelled
            part1, part2 = nx.algorithms.bipartite.sets(sub_g)

            # Ensure each partition contains only one type of node
            part1_labels = {sub_g.nodes[node].get("label") for node in part1}
            part2_labels = {sub_g.nodes[node].get("label") for node in part2}

            if part1_labels == {"data"} and part2_labels.issubset({"X", "Z"}):
                _, check_part = part1, part2

            elif part2_labels == {"data"} and part1_labels.issubset({"X", "Z"}):
                check_part, _ = part1, part2
            else:
                raise ValueError(
                    "Graph contains invalid edges among data or check nodes."
                )

            # Ensure that X ad Z nodes only share even number or data neighbors
            # This ensures that stabilizers commute
            neighbor_sets = {n: set(graph.neighbors(n)) for n in check_part}
            x_nodes = [
                n for n, info in graph.nodes(data=True) if info.get("label") == "X"
            ]
            z_nodes = [
                n for n, info in graph.nodes(data=True) if info.get("label") == "Z"
            ]

            if any(
                len(neighbor_sets[x_node] & neighbor_sets[z_node]) % 2
                for x_node, z_node in product(x_nodes, z_nodes)
            ):
                raise ValueError(
                    "X and Z check nodes share an odd number of data qubits. "
                    "This results in non-commuting stabilizers and "
                    "Tanner graph does not represent a valid stabilizer code."
                )

    def generate_graph_from_matrix(
        self, h_matrix: matrices.ParityCheckMatrix
    ) -> nx.Graph:
        """
        Generate Tanner graph from an input parity-check matrix. Note that, given our
        definition of TannerGraph, we only allow for ParityCheckMatrix objects
        describing CSS codes.

        Parameters
        ----------
        h_matrix : ParityCheckMatrix
            Parity-check matrix to generate the Tanner graph. Must be a CSS code.

        Returns
        -------
        g : nx.Graph
            Tanner graph describing the code, as a networkx graph.
        """

        # Verify input matrix is CSS
        if not h_matrix.is_css:
            raise ValueError("Parity-check matrix does not define a CSS code.")

        hx_matrix, hz_matrix = h_matrix.get_components()

        # Initialize graph
        g = nx.Graph()

        # Generate nodes
        data_nodes = [(i,) for i in range(h_matrix.n_datas)]
        x_nodes = [
            (i,) for i in range(h_matrix.n_datas, h_matrix.n_datas + h_matrix.n_xstabs)
        ]
        z_nodes = [
            (i,)
            for i in range(
                h_matrix.n_datas + h_matrix.n_xstabs,
                h_matrix.n_datas + h_matrix.n_stabs,
            )
        ]

        g.add_nodes_from(data_nodes, label="data")
        g.add_nodes_from(x_nodes, label="X")
        g.add_nodes_from(z_nodes, label="Z")

        # Generate edges
        for i, row in enumerate(hx_matrix.matrix):
            for j in np.where(row == 1)[0]:
                g.add_edge(data_nodes[j], x_nodes[i])

        for i, row in enumerate(hz_matrix.matrix):
            for j in np.where(row == 1)[0]:
                g.add_edge(data_nodes[j], z_nodes[i])

        return g

    def generate_graph_from_stabilizers(
        self, stabilizers: tuple[Stabilizer, ...]
    ) -> nx.Graph:
        """
        Generate Tanner graph from a tuple of stabilizers.

        Parameters
        ----------
        stabilizers : tuple[Stabilizer,...]
            Stabilizers to generate the Tanner graph.
        """

        # Ensure input is not empty
        if len(stabilizers) == 0:
            raise ValueError("Input tuple of stabilizers is empty.")

        # Verify list of stabilizers define a CSS code
        if not verify_css_code_stabilizers(stabilizers):
            raise ValueError(
                "TannerGraph generation requires input"
                " stabilizers to define a CSS code. Input"
                " contains non CSS stabilizers."
            )

        # Extract properties of datas and checks
        data_qubits = {qubit for stab in stabilizers for qubit in stab.data_qubits}

        x_qubits = {
            qubit
            for stab in stabilizers
            for qubit in stab.ancilla_qubits
            if set(stab.pauli) == {"X"}
        }

        z_qubits = {
            qubit
            for stab in stabilizers
            for qubit in stab.ancilla_qubits
            if set(stab.pauli) == {"Z"}
        }

        # Initialize graph
        t_graph = nx.Graph()

        t_graph.add_nodes_from(data_qubits, label="data")
        t_graph.add_nodes_from(x_qubits, label="X")
        t_graph.add_nodes_from(z_qubits, label="Z")

        # Generate edges
        edges = [
            (q, stab.ancilla_qubits[0])
            for stab in stabilizers
            for q in stab.data_qubits
        ]
        t_graph.add_edges_from(edges)

        return t_graph

    def to_stabilizers(self) -> list[Stabilizer]:
        """
        Convert Tanner graph to a list of stabilizers.
        """

        # Verify that nodes can be converted into coordinates
        all_nodes = list(self.graph.nodes)

        if not all(
            isinstance(t, tuple)
            and len(t) == len(all_nodes[0])
            and all(isinstance(e, int) for e in t)
            for t in all_nodes
        ):
            raise ValueError(
                "Nodes are not tuples of equal size and cannot be "
                "converted to list of stabilizers. Consider using "
                "relabel_graph() method, to re-index your graph and later"
                " convert to Stabilizer list."
            )

        stabilizers = []

        for x_node in self.x_nodes:
            data_nodes = list(self.graph.neighbors(x_node))
            stabilizer = Stabilizer(
                pauli="X" * len(data_nodes),
                data_qubits=[data_node + (0,) for data_node in data_nodes],
                ancilla_qubits=[x_node + (1,)],
            )
            stabilizers.append(stabilizer)

        for z_node in self.z_nodes:
            data_nodes = list(self.graph.neighbors(z_node))
            stabilizer = Stabilizer(
                pauli="Z" * len(data_nodes),
                data_qubits=[data_node + (0,) for data_node in data_nodes],
                ancilla_qubits=[z_node + (1,)],
            )
            stabilizers.append(stabilizer)

        return stabilizers

    def relabelled_graph(self) -> TannerGraph:
        """
        Relabel a Tanner graph to identify nodes with tuples of integers which can
        later be converted into Stabilizers. The nodes are a assigned a single-element
        tuple with an integer index. The data nodes are counted first, followed by the
        X check nodes and the Z check nodes. The old node labels are stored in the node
        attributes under the key 'original_node'.

        The relabelled graph is guaranteed to allow for conversion into Stabilizers
        through the `to_stabilizers()` method.

        Returns
        -------
        relabelled_tanner : TannerGraph
            Relabelled Tanner graph.
        """

        # Initialize graph
        t_graph = nx.Graph()

        # Add nodes
        data_nodes_info = {
            (i,): {"label": "data", "original_node": n}
            for i, n in enumerate(self.data_nodes)
        }
        data_nodes_dict = {v["original_node"]: k for k, v in data_nodes_info.items()}

        x_nodes_info = {
            (i + len(data_nodes_info),): {"label": "X", "original_node": n}
            for i, n in enumerate(self.x_nodes)
        }
        x_nodes_dict = {v["original_node"]: k for k, v in x_nodes_info.items()}

        z_nodes_info = {
            (i + len(data_nodes_info) + len(x_nodes_info),): {
                "label": "Z",
                "original_node": n,
            }
            for i, n in enumerate(self.z_nodes)
        }
        z_nodes_dict = {v["original_node"]: k for k, v in z_nodes_info.items()}

        t_graph.add_nodes_from(data_nodes_info.items())
        t_graph.add_nodes_from(x_nodes_info.items())
        t_graph.add_nodes_from(z_nodes_info.items())

        # Add edges
        for u, v in self.graph.edges():
            if self.graph.nodes[u]["label"] == "data":
                u, v = v, u
            if self.graph.nodes[u]["label"] == "X":
                t_graph.add_edge(data_nodes_dict[v], x_nodes_dict[u])
            else:
                t_graph.add_edge(data_nodes_dict[v], z_nodes_dict[u])

        # Construct new relabelled Tanner graph
        relabelled_tanner = TannerGraph(t_graph)

        return relabelled_tanner

    def get_components(
        self,
    ) -> tuple[ClassicalTannerGraph | None, ClassicalTannerGraph | None]:
        """
        Compute the X and Z components of the Tanner graph, associated with
        X and Z stabilizers, respectively.

        If the code contains only one type of stabilizer, the component of missing type
        will return None. The two of them cannot be None simultaneously.

        Returns
        -------
        t_graph_x, t_graph_z: tuple[ClassicalTannerGraph | None,
        ClassicalTannerGraph | None]
            Tuple containing the X and Z tanner graphs induced by the data nodes and the
            respective set of check nodes.
        """
        # Create subgraphs induced by the selected nodes and convert
        # into Classical Tanner Graphs
        if len(self.x_nodes) == 0:
            z_component = self.graph.copy().subgraph(self.data_nodes + self.z_nodes)
            t_graph_z = ClassicalTannerGraph(z_component)
            t_graph_x = None

        elif len(self.z_nodes) == 0:
            x_component = self.graph.copy().subgraph(self.data_nodes + self.x_nodes)
            t_graph_x = ClassicalTannerGraph(x_component)
            t_graph_z = None

        else:
            x_component = self.graph.copy().subgraph(self.data_nodes + self.x_nodes)
            z_component = self.graph.copy().subgraph(self.data_nodes + self.z_nodes)
            t_graph_x = ClassicalTannerGraph(x_component)
            t_graph_z = ClassicalTannerGraph(z_component)

        return t_graph_x, t_graph_z

    def __eq__(self, other: TannerGraph) -> bool:
        """
        Check if two Tanner graphs are equal. Two Tanner graphs are equal if they have
        the same graphs and same lists of data, x and z nodes.

        Parameters
        ----------
        other : TannerGraph
            TannerGraph to compare with.

        Returns
        -------
        bool
            True if the two Tanner graphs are equal, False otherwise.
        """
        if not isinstance(other, TannerGraph):
            raise TypeError("Comparison is only supported between TannerGraph objects.")

        def are_node_attributes_not_the_same():
            return (
                set(self.data_nodes) != set(other.data_nodes)
                or set(self.x_nodes) != set(other.x_nodes)
                or set(self.z_nodes) != set(other.z_nodes)
            )

        def are_nodes_not_the_same():
            return self.graph.nodes(data=True) != other.graph.nodes(data=True)

        # Check edges by sorting the tuples to ensure order does not matter
        def are_edges_not_the_same():
            return set(tuple(sorted(edge)) for edge in self.graph.edges()) != set(
                tuple(sorted(edge)) for edge in other.graph.edges()
            )

        return not (
            are_node_attributes_not_the_same()
            or are_nodes_not_the_same()
            or are_edges_not_the_same()
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from copy import deepcopy

from .operations.base_operation import Operation
from .operations.datamanipulation_operation import CreatePauliFrame, RecordPauliFrame
from .operations.resize_operation import AddQubit, DeleteQubit
from .classicalreg import ClassicalRegister
from .moment_queue import MomentQueue
from .tableau import Tableau
from .data_store import DataStore
from .invoker import Invoker
from .exceptions import EngineRunError


class Engine:  # pylint: disable=too-many-instance-attributes
    """
    The Engine Object contains all the components that make up the Stabilizer
    Simulator. These include the MomentQueue, the Tableau and the
    DataStore.

    This Object facilitates the interaction between all of these internal
    components with each other.
    """

    def __init__(
        self,
        input_operations: list[Operation],
        nqubits: int,
        seed: int = None,
        shots: int = 1,
        parallelize: bool = False,
    ):
        self.input_operations = input_operations
        self._validate_operations(input_operations, nqubits)
        self.moment_queue = MomentQueue(input_operations, parallelize)
        self.nqubits = nqubits
        self.seed = seed
        self.shots = shots

        # initialize clifford simulator
        self._setup_clifford_simulator()

    # pylint: disable=too-many-branches
    def _validate_operations(self, input_operations: list[Operation], nqubits: int):
        """Validate input operations."""
        created_pfs = []
        recorded_pfs = []
        for operation in input_operations:
            if not isinstance(operation, Operation):
                raise TypeError(
                    "Input operations must be of type Operation or a subclass of "
                    "Operation."
                )
            if isinstance(operation, CreatePauliFrame):
                created_pfs.append(operation.pauli_frame)
            if isinstance(operation, RecordPauliFrame):
                recorded_pfs.append(operation.pauli_frame)

        # Check that all recorded PF have been created first
        for pauli_frame in recorded_pfs:
            if pauli_frame not in created_pfs:
                raise ValueError(
                    "RecordPauliFrame operations must be preceded by a "
                    "CreatePauliFrame operation."
                )

        # Check that all created PF have a unique id
        compare_pfs = deepcopy(created_pfs)
        for pauli_frame in created_pfs:
            compare_pfs.pop(0)  # remove index 0 to avoid comparing with itself
            for other_pauli_frame in compare_pfs:
                if pauli_frame.id == other_pauli_frame.id:
                    raise ValueError(
                        "CreatePauliFrame operations must have a unique PauliFrame id."
                    )

        # Check that PauliFrames have the right size
        compare_nqubits = nqubits
        for input_operation in input_operations:
            if isinstance(input_operation, AddQubit):
                compare_nqubits += 1
            if isinstance(input_operation, DeleteQubit):
                compare_nqubits -= 1
            if isinstance(input_operation, CreatePauliFrame):
                if len(input_operation.pauli_frame.x) != compare_nqubits:
                    raise ValueError(
                        f"Wrong size for the PauliFrame "
                        f"{input_operation.pauli_frame.id}. It has size "
                        f"{len(input_operation.pauli_frame.x)}. It must have "
                        "the same length as the number of qubits in the system "
                        f"({compare_nqubits}). Make sure that you take into "
                        "account resize operations."
                    )

    def _setup_clifford_simulator(self):
        """Set up the clifford simulator."""
        # Setup tableau
        # Initialize using nqubits attribute
        self.tableau_w_scratch = Tableau(nqubits=self.nqubits, seed=self.seed)

        # Setup PauliFrame lists, default empty
        self.pauli_frames_forward = []
        self.pauli_frames_backward = []

        # Setup Classical Register, default empty.
        self.registry: dict[str, ClassicalRegister] = {}

        # Initialize DataStore and Invoker objects
        self.data_store = DataStore()
        self.invoker = Invoker(
            self.tableau_w_scratch,
            self.pauli_frames_forward,
            self.pauli_frames_backward,
            self.data_store,
            self.registry,
        )

    def run(self):
        """Run the clifford simulator."""
        # get one by one all the moments in both directions
        for forward_moment, backward_moment in self.moment_queue.moment_generators:
            # Forward propagation
            output_bool_tab = self.invoker.transform_tab(forward_moment)
            output_bool_pf = self.invoker.transform_pf(forward_moment)
            # Check for errors when applying the forward transformations
            if not output_bool_tab:
                raise EngineRunError(
                    """An exception has occured while trying to apply a
                    transformation to the Tableau."""
                )
            if not output_bool_pf:
                raise EngineRunError(
                    """An exception has occured while trying to apply a forward
                    transformation to the PauliFrame."""
                )
            # Backward propagation
            output_bool_pf_back = self.invoker.transform_pf_back(backward_moment)
            # Check for errors when applying the backward transformations
            if not output_bool_pf_back:
                raise EngineRunError(
                    """An exception has occured while trying to apply a backward
                    transformation to the PauliFrame."""
                )

        # reset the moment queue
        self.moment_queue.reset_queue()

    @property
    def stabilizer_set(self) -> set[str]:
        """
        Returns the stabilizer set in string format.
        """
        return self.tableau_w_scratch.stabilizer_set

    @property
    def stabilizer_set_sparse_format(self) -> list[dict]:
        """
        Returns the stabilizer set in sparse format.
        """
        return self.tableau_w_scratch.stabilizer_set_sparse_format



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from dataclasses import dataclass, field
import uuid
from typing import Optional

from .base_operation import Operation, OpType
from .controlled_operation import has_ccontrol


@dataclass
@has_ccontrol
class MeasurementOperation(Operation):
    """
    Operations of this type perform measurements on qubits within the Engine during
    runtime.
    """

    operation_type: str = field(default=OpType.MEASUREMENT, init=False)
    target_qubit: int


@dataclass
class Measurement(MeasurementOperation):
    """
    Creates a measurement operation for a single qubit.
    By default, the output of the measurement is saved to the DataStore. If the user
    wants to write the output onto the classical register, the classical register
    needs to be specified with reg_name and (bit_order or bit_id).

    Parameters
    ----------
    label: str
        The ID of the Measurement Operation. Automatically generated by Default.
    reg_name: Optional[str]
        The name of the classical register to write the measurement output to.
    bit_order: Optional[int]
        The ordering of the bit within the classical register to write the measurement
        output to.
    bit_id: Optional[str]
        The bit ID of the bit within the classical register to write the measurement
        output to.
    basis: str
        The basis of the measurement. Default is Z.
    bias: float
        The bias of the measurement. Default is 0.5.
    """

    name: str = field(default="Measurement", init=False)
    label: str = field(default_factory=lambda: str(uuid.uuid4()), init=True)
    reg_name: Optional[str] = field(default=None)
    bit_order: Optional[int] = field(default=None)
    bit_id: Optional[str] = field(default=None)

    # Parameters for the measurement operation
    basis: str = field(default="Z")
    bias: float = field(default=0.5)

    def __post_init__(self):
        # Checks only trigger if a reg_name is specified.
        if self.reg_name is not None:
            if self.bit_order is None and self.bit_id is None:
                raise ValueError(
                    "Either the bit order or the bit id of the classical bit has to be"
                    "specified."
                )
            if self.bit_order is not None and self.bit_id is not None:
                raise ValueError(
                    "Both bit_order and bit_id cannot be specified together. Only input"
                    " 1 parameter, not both."
                )


@dataclass
class Reset(MeasurementOperation):
    """
    Resets a qubit to a specific cardinal state. By default, the qubit is reset to
    the `|0>` state.

    Parameters
    ----------
    state: str
        The state to which the qubit is reset. The state can be one of the following:
        0, 1, +, -, +i, -i. Default is 0.
    """

    name: str = field(default="Reset", init=False)
    label: str = field(default_factory=lambda: str(uuid.uuid4()), init=True)
    state: str = "0"

    def __post_init__(self):
        if self.state not in ["0", "1", "+", "-", "+i", "-i"]:
            raise ValueError(
                "The state of the reset operation must be either 0, 1, "
                "+, -, +i, or -i."
            )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from enum import Enum
from functools import cached_property, partial
from typing import Any

from pydantic import BaseModel, Field

from ..eka import Circuit
from ..interpreter import InterpretationStep


from .converter import Converter
from .eka_to_pennylane_converter import EkaToPennylaneConverter
from .eka_to_qasm_converter import EkaToQasmConverter
from .eka_to_guppylang_converter import EkaToGuppylangConverter
from .eka_to_stim_converter import EkaToStimConverter
from .eka_to_mimiq_converter import EkaToMimiqConverter
from .eka_to_cudaq_converter import EkaToCudaqConverter


class TargetLanguage(Enum):
    """Enumeration of the available converters."""

    MIMIQ = EkaToMimiqConverter
    OPEN_QASM3 = EkaToQasmConverter
    GUPPYLANG = EkaToGuppylangConverter
    STIM = EkaToStimConverter
    PENNYLANE = partial(EkaToPennylaneConverter, is_catalyst=False)
    PENNYLANE_CATALYST = partial(EkaToPennylaneConverter, is_catalyst=True)
    CUDAQ = EkaToCudaqConverter


class Executor(BaseModel):
    """
    Factory class for executing quantum circuits using different target languages.

    Parameters
    ----------
    target_language : TargetLanguage
        The target language/platform for circuit execution
    """

    target_language: TargetLanguage = Field(
        description="Target language for circuit conversion and execution"
    )

    # Pydantic configuration
    model_config = {
        "arbitrary_types_allowed": True,  # Allow custom types like Converter
    }

    @cached_property
    def converter(self) -> Converter:
        """Return the converter for the target language."""
        return self.target_language.value()  # pylint: disable=no-member

    def export_eka(self, input_data: InterpretationStep) -> Any:
        """
        Factory function that selects the right converter and runs it.

        Parameters
        ----------
        input_data : InterpretationStep
            The EKA interpretation step to convert

        Returns
        -------
        Any
            The converted output in the target language format
        """
        return self.converter.convert(input_data)

    def export_eka_circuit(self, input_data: Circuit) -> Any:
        """
        Export the EKA circuit to the target language format.

        Parameters
        ----------
        input_data : Circuit
            The EKA circuit to convert

        Returns
        -------
        Any
            The converted output in the target language format
        """
        return self.converter.convert_circuit(input_data)

    def parse_outcome(self, run_output: Any) -> dict[str, int | list[int]]:
        """
        Function to parse the measurement outcomes from the target language run output.

        Parameters
        ----------
        run_output : Any
            The output from running the converted circuit

        Returns
        -------
        dict[str, int | list[int]]
            Parsed measurement outcomes as a dictionary mapping channels labels to
            outcomes or lists of outcomes (for multiple shots)
        """
        return self.converter.parse_target_run_outcome(run_output)



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

import numpy as np
import plotly.graph_objs as go
import networkx as nx

from loom.eka import Circuit


def hex_to_rgb(hex_color: str) -> list[int]:
    """
    Converts a hexadecimal color code to a list of RGB values between 0 and 255.

    Parameters
    ----------
    hex_color : str
        The hexadecimal color code (e.g., "#RRGGBB").

    Returns
    -------
    list[int]
        RGB values between 0 and 255 representing the color.
    """
    # Remove "#" if present
    if hex_color.startswith("#"):
        hex_color = hex_color[1:]

    # Convert hex color to RGB components
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)

    return [r, g, b]


def rgb_to_hex(rgb_color: list[int]) -> str:
    """
    Converts a color from RGB format to HEX format.

    Parameters
    ----------
    rgb_color : list[int]
        RGB values between 0 and 255 representing the color.

    Returns
    -------
    str
        The hexadecimal color code (e.g., "#RRGGBB").
    """
    return f"#{rgb_color[0]:02x}{rgb_color[1]:02x}{rgb_color[2]:02x}"


def average_color_hex(color_list: list[str]) -> str:
    """
    Calculate the average color from a list of colors in HEX format.

    Parameters
    ----------
    color_list : list[str]
        List of hexadecimal colors

    Returns
    -------
    str
        The average color in HEX format
    """
    rgb_colors = [hex_to_rgb(color) for color in color_list]
    avg_rgb = tuple(
        int(sum(color_channel) / len(color_list))
        for color_channel in zip(*rgb_colors, strict=True)
    )
    return rgb_to_hex(avg_rgb)


def change_color_brightness(hex_color: str, factor: float) -> str:
    """
    Change the brightness of a color by a given factor.

    Parameters
    ----------
    hex_color : str
        The hexadecimal color code (e.g., "#RRGGBB").
    factor : float
        The factor by which the brightness of the color should be multiplied.

    Returns
    -------
    str
        The new color in HEX format.
    """
    if factor < 0:
        raise ValueError("Factor must be greater than or equal to 0.")
    rgb_color = hex_to_rgb(hex_color)
    new_rgb_color = [min(int(np.round(color * factor)), 255) for color in rgb_color]
    new_hex_color = rgb_to_hex(new_rgb_color)
    return new_hex_color


def get_font_color(background_color: str) -> str:
    """
    Determine font color based on background color. The font color is white for dark
    backgrounds and gray or black for bright backgrounds.

    Parameters
    ----------
    background_color : str
        Background color in hexadecimal format (e.g., '#RRGGBB')

    Returns
    -------
    str
        Font color for a good contrast with the background color
    """
    # Convert hexadecimal color to RGB
    r = int(background_color[1:3], 16)
    g = int(background_color[3:5], 16)
    b = int(background_color[5:], 16)

    # Calculate luminance (YIQ formula)
    luminance = (r * 0.299 + g * 0.587 + b * 0.114) / 255

    # Determine font color based on luminance
    if luminance > 0.7:
        return "#000000"
    if luminance > 0.5:
        return "#222222"
    return "#ffffff"


def point_in_polygon(
    x: float,
    y: float,
    polygon: list[tuple[float, float]],
) -> bool:
    """
    Check if a point (x, y) is inside a polygon defined by its corners. This
    algorithm is based on the ray casting algorithm. Note that the behaviour for points
    on the edges or very close to them is undefined.

    Parameters
    ----------
    x : float
        x coordinate of the point to check
    y : float
        y coordinate of the point to check
    polygon : list[tuple[float, float]]
        List of tuples representing the corners of the polygon

    Returns
    -------
    bool
        True if the point is inside the polygon, False otherwise
    """
    # How the ray casting algorithm works:
    # We consider a horizontal line starting from the point to check and going to the
    # right. We count how many times this line intersects with the edges of the polygon.
    # If the number of intersections is odd, the point is inside the polygon. If it is
    # even, the point is outside the polygon.
    # In the code, instead of an integer counter, we use a boolean variable is_inside
    # which is toggled every time the line intersects with an edge.
    is_inside = False
    for i, corner in enumerate(polygon):  # Iterate over each edge of the polygon
        # corner = polygon[i] is the current corner
        # polygon[j] is the next corner of the polygon
        j = (i + 1) % len(polygon)
        next_corner = polygon[j]

        # Check if the y coordinate of the point is between the upper and lower y
        # coordinates of the edge
        point_is_in_y_range = (corner[1] > y) != (next_corner[1] > y)
        # If the point is not in the y range, we can skip this edge since there cannot
        # be an intersection
        if not point_is_in_y_range:
            continue

        # Check if the x coordinate of the point is to the left of the point on the edge
        # which has the same y coordinate as the given point
        if abs(next_corner[1] - corner[1]) > 1e-9:  # Avoid division by zero
            inverse_slope = (next_corner[0] - corner[0]) / (next_corner[1] - corner[1])
            # Inverse slope of line going from (x1, y1) to (x2, y2) is
            # (x2 - x1) / (y2 - y1)

            point_is_left_of_edge = x < corner[0] + inverse_slope * (y - corner[1])

            # Check if the horizontal line going to the right intersects with the edge
            if point_is_left_of_edge:
                is_inside = not is_inside
        else:
            # This is the case where the edge is practically horizontal and the point
            # might or might not lie on the edge. As stated in the docstring, the
            # behaviour for points on the edges or very close to them is undefined.
            # Therefore, we just skip this edge.
            continue

    return is_inside


def interpolate_values(
    point: tuple[float, float],
    interpolation_points: list[tuple[float, float]],
    interpolation_values: list[float] | list[list[float]],
    interpolation_power: float | None = None,
    min_distance: float | None = 1e-3,
) -> np.ndarray:
    """
    Perform interpolation of the values in `interpolation_values` which are
    associated to the points in `interpolation_points`, based on the distances to a
    given point. The values in `interpolation_values` are either floats or lists of
    floats. If they are a list, the interpolation is done element- wise. E.g. every
    interpolation point could have a tuple (r, g, b) of three integers, representing an
    RGB color. Then the weighted average of these RGB tuples is calculated with weights
    determined by the distances to these points.

    Parameters
    ----------
    point : tuple[float, float]
        Tuple representing the coordinates of the given point
    interpolation_points : list[tuple[float, float]]
        List of points at which the values in interpolation_values are given
    interpolation_values : list[float] | list[list[float]]
        List of values associated to the points in interpolation_points. The values can
        be floats or lists of floats. If they are lists, the interpolation is done
        element-wise.
    interpolation_power : float | None
        The interpolation_power parameter determines how rapidly the interpolated value
        changes with increasing distance. The weights with which the values of the n
        points are interpolated are given as 1 / distance**interpolation_power.
        The default value is 2.
    min_distance : float | None
        Points which are closer to one of the given points than min_distance will get a
        value equal to the value of this point. This is to prevent division by zero when
        calculating the weights.

    Returns
    -------
    np.ndarray
        Interpolated value(s) based on the distances from the given point. The length of
        the array is equal to the length of the values in interpolation_values.
    """
    if interpolation_power is None:
        interpolation_power = 2

    # Make sure the number of interpolation points and interpolation values is the same
    if len(interpolation_points) != len(interpolation_values):
        raise ValueError(
            "The number of interpolation points and interpolation values must be the "
            "same."
        )

    # Make sure every element of interpolation_values is a list.
    # If it is a single value, convert it to a list of length 1.
    for i, val in enumerate(interpolation_values):
        if not isinstance(val, (tuple, list)):
            interpolation_values[i] = [val]

    # Check if all interpolation_values have the same dimension
    if any(len(val) != len(interpolation_values[0]) for val in interpolation_values):
        raise ValueError(
            "All values in interpolation_values must have the same dimension."
        )

    # Check that all points in interpolation_points are tuples of length 2
    if any(
        not (isinstance(point, tuple)) or len(point) != 2
        for point in interpolation_points
    ):
        raise ValueError(
            "All points in interpolation_points must be tuples of length 2."
        )

    total_value = np.array([0] * len(interpolation_values[0]), dtype=float)
    total_weight = 0

    # Calculate the weighted average of the values of the n points
    for pt, value in zip(interpolation_points, interpolation_values, strict=True):
        distance = ((point[0] - pt[0]) ** 2 + (point[1] - pt[1]) ** 2) ** 0.5
        if distance < min_distance:
            return value

        weight = 1 / distance**interpolation_power
        total_value += np.array(value) * weight
        total_weight += weight

    return total_value / total_weight


def center_of_points(points: list[tuple[float, float]]) -> np.ndarray:
    """Calculate the center of a list of points."""
    x_coords = [point[0] for point in points]
    y_coords = [point[1] for point in points]
    return np.array([np.mean(x_coords), np.mean(y_coords)])


def center_of_scatter_plot(scatter_plot: go.Scatter) -> np.ndarray:
    """Calculate the center of the points contained in a scatter plot."""
    coords = np.stack((scatter_plot.x, scatter_plot.y), axis=1)
    return np.average(coords, axis=0)


def get_angle_from_x_axis(
    point1: tuple[float, float], point2: tuple[float, float]
) -> float:
    """
    Calculate the angle between the vector from point1 to point2 and the x-axis.

    E.g. for point1 == [0,0] and point2 == [1,0], the angle is 0. For point1 == [0,0]
    and point2 == [0,1], the angle is pi/2 (90 degree).
    """
    delta = np.array(point2) - np.array(point1)
    return np.arctan2(delta[1], delta[0])


def order_points_counterclockwise(
    points: list[tuple[float, float, list[any]]],
) -> list[tuple[float, float, list[any]]]:
    """
    Order the provided list of points counterclockwise around the center point.

    Parameters
    ----------
    points : list[tuple[float, float, list[any]]]
        List of points to be ordered. Every element in the list is a tuple which
        contains the x and y coordinate of the point as the first two elements. The
        third element is a list which can store additional metadata.

    Returns
    -------
    list[tuple[float, float, list[any]]]
        List of points ordered counterclockwise around the center point of the provided
        list. The list still contains the original metadata. As an additional metadata,
        the angle for the every point was added to its respective list.
    """
    center_point = center_of_points(points)
    points_with_angles = [list(point) for point in points]
    for point, point_with_angle in zip(points, points_with_angles, strict=True):
        # If the points have no additional metadata, add an empty list to add the angle
        if len(point_with_angle) == 2:
            point_with_angle.append([])

        point_with_angle[2].append(
            get_angle_from_x_axis(center_point, np.array([point[0], point[1]]))
        )
    sorted_points = sorted(points_with_angles, key=lambda x: x[2][-1])
    return sorted_points


# pylint: disable=too-many-arguments, too-many-positional-arguments
def draw_half_circle(
    center: list[float],
    r: float,
    direction: float | None = 0,
    name: str | None = "",
    text: str | None = "",
    fillcolor: str | None = "white",
    line: dict | None = None,
    showlegend: bool = True,
    legendgroup: str | None = None,
) -> go.Scatter:
    """
    Generate a scatter plot for a half circle.

    Parameters
    ----------
    center : list[float]
        Center point around which the half circle should be drawn
    r : float
        Radius of the half circle
    direction : float | None
        Direction in which the half circle should point. E.g.,

        - If direction == 0, half circle points upwards.
        - If direction == pi/2, half circle points right.
        - If direction == - pi/2, half circle points left.
    name : str | None
        Name parameter of the scatter plot
    text : str | None
        Text parameter of the scatter plot
    fillcolor : str | None
        Fill color of the half circle
    line : dict | None
        Line parameter of the scatter plot
    showlegend : bool, optional
        If True (the default), the stabilizers are shown in the legend.
    legendgroup: str | None
        Legend group to which the stabilizers belong.

    Returns
    -------
    go.Scatter
        Scatter plot with the half circle
    """
    if line is None:
        line = {}
    theta = np.linspace(0, np.pi, 100)

    # Calculate x and y coordinates for half circle
    x = center[0] + r * np.cos(-direction + theta)
    y = center[1] + r * np.sin(-direction + theta)

    # Add the starting point to have a closed shape
    x = np.append(x, x[0])
    y = np.append(y, y[0])

    # Add half circle as scatter plot
    return go.Scatter(
        x=x,
        y=y,
        mode="lines",
        name=name,
        fill="toself",
        fillcolor=fillcolor,
        line=line,
        text=text,
        hoverinfo="text",
        showlegend=showlegend,
        legendgroup=legendgroup,
    )


def get_label_for_circuit(circ: Circuit) -> str:
    """
    Generate a label for a circuit. If the circuit has subcircuits, the label is
    simply the name attribute of the circuit. If the circuit has no subcircuits, the
    label contains the circuit's name (specifying the gate) and the labels of the
    channels that are involved in the circuit.

    Parameters
    ----------
    circ : Circuit
        Circuit for which the label should be generated

    Returns
    -------
    str
        Label for the circuit
    """
    if len(circ.circuit) == 0:
        qb_list_str = ",".join(ch.label for ch in circ.channels)
        return f"{circ.name}({qb_list_str})"

    return circ.name


def convert_circuit_to_nx_graph(circ: Circuit) -> tuple[nx.DiGraph, list[str]]:
    """
    Construct a NetworkX directed graph (DiGraph) from a circuit. The nodes of the graph
    are all the subcircuits contained in the circuit object. The edges are directed from
    every circuit to its subcircuits.

    Parameters
    ----------
    circ : Circuit
        Circuit from which the graph should be constructed

    Returns
    -------
    nx.DiGraph: Directed graph representing the circuit
    list[str]: List of labels for the nodes in BFS traversal order
    """
    graph = nx.DiGraph()
    labels_nodes = []
    graph.add_node(circ.id)
    labels_nodes.append(get_label_for_circuit(circ))

    # This is the Breadth First Search (BFS) traversal of a tree:
    # https://en.wikipedia.org/wiki/Tree_traversal#Breadth-first_search
    queue = [circ]
    while len(queue) > 0:
        next_circuit = queue.pop(0)

        for c in next_circuit.circuit:  # Iterate over all subcircuits
            if len(c) > 0:  # Skip empty tuples
                graph.add_node(c[0].id)  # Add node to graph
                graph.add_edge(
                    next_circuit.id, c[0].id
                )  # Add edge from parent circuit to this subcircuit
                # Add label for the plot
                labels_nodes.append(get_label_for_circuit(c[0]))

                # If this subcircuit contains more subcircuits which have to be added to
                # the graph, add it to the queue so that it will be processed later in
                # the right order
                if len(c[0].circuit) > 0:
                    queue.append(c[0])

    return graph, labels_nodes



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

# pylint: disable=duplicate-code
from loom.eka import Circuit, Stabilizer
from loom.eka.utilities import DiagonalDirection, Orientation
from loom.interpreter import InterpretationStep

from .y_wall_out_utilities import (
    generate_teleportation_finalization_circuit_with_updates,
    find_stabilizer_position_in_final_block,
    qubit_mapper,
    move_logicals_and_append_evolution_and_updates,
)
from ..move_block import (
    DetailedSchedule,
    direction_to_coord,
    find_swap_then_qec_qubit_initializations,
    generate_syndrome_measurement_circuit_and_cbits,
    generate_and_append_block_syndromes_and_detectors,
)
from ..utilities import FourBodySchedule, find_relative_diagonal_direction
from ...code_factory import RotatedSurfaceCode


def y_wall_out_recombination_swap_then_qec(
    # pylint: disable=too-many-locals, too-many-arguments, too-many-positional-arguments
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    initial_block: RotatedSurfaceCode,
    qubits_to_idle: list[tuple[int, ...]],
    qubits_measured: list[tuple[int, ...]],
    qubits_to_had: list[tuple[int, ...]],
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Apply the y_wall_out recombination using the swap-then-qec method. The initial
    block that is missing the y_wall qubits is recombined into a new, compact block
    on the dual lattice.

    - A) Geometry

      - A.1) Find idle and hadamard side stabilizers and their directions along with \
            wall stabilizers

    - B) New stabilizers with evolution and updates

    - C) Logical operators with evolution and updates

    - D) Define new Block

    - E) Circuit generation

      - E.1) Find wall, idle and hadamard side qubits to initialize (ancillas, data, \
        and teleportation pairs)
      - E.2) Generate initialization circuit
      - E.3) Generate CNOT circuit
      - E.4) Generate teleportation finalization circuit and syndrome measurement \
        circuit
      - E.5) Assemble and append the full circuit to the interpretation step

    - F) Update Block history and evolution
    
    - G) Generate and append Block syndromes and detectors

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The interpretation step.
    block: RotatedSurfaceCode
        The block to which the operation will be applied.
    initial_block: RotatedSurfaceCode
        The initial block before the y_wall_out operation. This is needed to find the
        wall data qubits to initialize.
    qubits_to_idle: list[tuple[int, ...]]
        The list of qubits to idle.
    qubits_measured: list[tuple[int, ...]]
        The list of qubits to measure.
    qubits_to_had: list[tuple[int, ...]]
        The list of qubits to Hadamard.
    same_timeslice: bool
        Whether to append the generated circuit to the same timeslice as the last
        circuit in the interpretation step.
    debug_mode: bool
        Whether to enable debug mode.
    """

    # A.) Geometry
    # A.1) Find idle and hadamard side stabilizers and their directions along with wall
    # stabilizers
    (
        idle_side_stabilizers,
        had_side_stabilizers,
        wall_stabilizers,
        idle_side_directions,
        had_side_directions,
    ) = get_idle_hadamard_info(
        block,
        qubits_to_had,
        qubits_to_idle,
    )

    # B.) New stabilizers with evolution and updates
    # Find new stabilizers after recombination and append their evolution and updates
    new_stabilizers = find_new_stabilizers_and_append_evolution_and_updates(
        interpretation_step,
        qubits_to_idle,
        idle_side_stabilizers,
        had_side_stabilizers,
        wall_stabilizers,
        idle_side_directions,
        had_side_directions,
    )

    # C.) Logical operators with evolution and updates
    # Move logical operators and append their evolution and updates
    new_logical_x, new_logical_z = move_logicals_and_append_evolution_and_updates(
        interpretation_step, block, idle_side_directions
    )

    # D.) Define new Block
    recombined_block = RotatedSurfaceCode(
        stabilizers=new_stabilizers,
        logical_x_operators=(new_logical_x,),
        logical_z_operators=(new_logical_z,),
        unique_label=block.unique_label,
        skip_validation=not debug_mode,
    )

    # E.) Circuit generation
    # E.1) Find wall, idle and hadamard side qubits to initialize (ancillas, data, and
    # teleportation pairs)
    # Wall qubits to initialize
    wall_data_qubits_to_init = get_wall_data_qubit_init(initial_block, qubits_measured)
    # Idle side qubits to initialize
    (
        anc_qubits_to_init_idle,
        data_qubits_to_init_idle,
        teleportation_qubit_pairs_idle,
    ) = find_swap_then_qec_qubit_initializations(
        idle_side_stabilizers,
        idle_side_directions,
    )
    # Hadamard side qubits to initialize
    anc_qubits_to_init_had, data_qubits_to_init_had, teleportation_qubit_pairs_had = (
        find_swap_then_qec_qubit_initializations(
            had_side_stabilizers,
            had_side_directions,
        )
    )

    # E.2) Generate initialization circuit
    first_swap_then_qec_reset_circuit = Circuit(
        name="Initialization of qubits for first swap-then-qec",
        circuit=[
            [
                Circuit(
                    f"reset_{'0' if pauli == 'Z' else '+'}",
                    channels=[interpretation_step.get_channel_MUT(q)],
                )
                for pauli in ["X", "Z"]
                for q in wall_data_qubits_to_init[pauli]
                + data_qubits_to_init_idle[pauli]
                + data_qubits_to_init_had[pauli]
                + anc_qubits_to_init_had[pauli]
                + anc_qubits_to_init_idle[pauli]
            ]
        ],
    )

    # E.3) Generate CNOT circuit
    first_swap_then_qec_cnots_circuit = (
        get_final_block_syndrome_measurement_first_swap_then_qec_cnots_circuit(
            interpretation_step,
            recombined_block,
            qubits_to_idle,
            qubits_to_had,
            idle_side_directions,
            had_side_directions,
            anc_qubits_to_init_idle,
            anc_qubits_to_init_had,
        )
    )
    # E.4) Generate teleportation finalization circuit and syndrome measurement circuit
    # Teleportation finalization circuit
    all_ancillas_to_init = {
        "X": anc_qubits_to_init_idle["X"] + anc_qubits_to_init_had["X"],
        "Z": anc_qubits_to_init_idle["Z"] + anc_qubits_to_init_had["Z"],
    }
    first_swap_then_qec_teleportation_finalization_circuit = (
        generate_teleportation_finalization_circuit_with_updates(
            interpretation_step,
            recombined_block,
            all_ancillas_to_init,
            teleportation_qubit_pairs_idle + teleportation_qubit_pairs_had,
        )
    )
    # Syndrome measurement circuit and classical bits
    first_swap_then_qec_measurement_circuit, first_swap_then_qec_cbits = (
        generate_syndrome_measurement_circuit_and_cbits(
            interpretation_step,
            recombined_block,
        )
    )

    # E.5) Assemble and append the full circuit to the interpretation step
    interpretation_step.append_circuit_MUT(
        Circuit(
            name="swap-then-qec to recombine block",
            circuit=Circuit.construct_padded_circuit_time_sequence(
                (
                    (first_swap_then_qec_reset_circuit,),
                    (first_swap_then_qec_cnots_circuit,),
                    (
                        first_swap_then_qec_measurement_circuit,
                        first_swap_then_qec_teleportation_finalization_circuit,
                    ),
                ),
            ),
        ),
        same_timeslice=same_timeslice,
    )

    # F.) Update Block history and evolution
    interpretation_step.update_block_history_and_evolution_MUT(
        (recombined_block,), (block,)
    )

    # G.) Generate and append Block syndromes and detectors
    generate_and_append_block_syndromes_and_detectors(
        interpretation_step, recombined_block, first_swap_then_qec_cbits
    )

    return interpretation_step


def get_idle_hadamard_info(
    block: RotatedSurfaceCode,
    qubits_to_had: list[tuple[int, ...]],
    qubits_to_idle: list[tuple[int, ...]],
) -> tuple[
    list[Stabilizer],
    list[Stabilizer],
    list[Stabilizer],
    DiagonalDirection,
    DiagonalDirection,
]:
    """
    Find the stabilizers associated with the idling and hadamard side of the block
    along with their directions.

    Parameters
    ----------
    block: RotatedSurfaceCode
        The block to which the operation will be applied.
    qubits_to_had: list[tuple[int, ...]]
        The list of qubits to Hadamard.
    qubits_to_idle: list[tuple[int, ...]]
        The list of qubits to idle.

    Returns
    -------
    list[Stabilizer]
        The stabilizers associated with the idling side.
    list[Stabilizer]
        The stabilizers associated with the hadamard side AFTER the transversal
        Hadamard.
    DiagonalDirection
        The directions of movement of the idling side.
    DiagonalDirection
        The directions of movement the hadamard side.
    """

    # Deduce geometric parameters
    block_orientation = block.orientation
    is_top_left_bulk_stab_x = block.upper_left_4body_stabilizer.pauli_type == "X"

    # Find idle side stabilizers
    idle_side_stabilizers = [
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).issubset(qubits_to_idle)
    ]

    # Find had side stabilizers along with their status after the Hadamard operation.
    # We need to find the stabilizers and then change the pauli type of the stabilizer.
    had_side_stabilizers = [
        stab
        for stab in block.stabilizers
        if set(stab.data_qubits).issubset(qubits_to_had)
    ]

    wall_stabilizers = [
        stab
        for stab in block.stabilizers
        if stab not in idle_side_stabilizers + had_side_stabilizers
    ]

    # Find idle and hadamard side directions
    match (block_orientation, is_top_left_bulk_stab_x):
        case (Orientation.VERTICAL, False):
            idle_side_directions = DiagonalDirection.BOTTOM_RIGHT
            had_side_directions = DiagonalDirection.TOP_RIGHT
        case (Orientation.VERTICAL, True):
            idle_side_directions = DiagonalDirection.BOTTOM_LEFT
            had_side_directions = DiagonalDirection.TOP_LEFT
        case (Orientation.HORIZONTAL, True):
            idle_side_directions = DiagonalDirection.TOP_RIGHT
            had_side_directions = DiagonalDirection.TOP_LEFT
        case (Orientation.HORIZONTAL, False):
            idle_side_directions = DiagonalDirection.BOTTOM_RIGHT
            had_side_directions = DiagonalDirection.BOTTOM_LEFT

    return (
        idle_side_stabilizers,
        had_side_stabilizers,
        wall_stabilizers,
        idle_side_directions,
        had_side_directions,
    )


def find_new_stabilizers_and_append_evolution_and_updates(
    interpretation_step: InterpretationStep,
    qubits_to_idle: list[tuple[int, ...]],
    idle_side_stabilizers: list[Stabilizer],
    had_side_stabilizers: list[Stabilizer],
    wall_stabilizers: list[Stabilizer],
    idle_side_directions: DiagonalDirection,
    had_side_directions: DiagonalDirection,
) -> list[Stabilizer]:
    """
    Find the new stabilizers after the recombination of the block and
    append their evolution and updates to the interpretation step.

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.
    qubits_to_idle: list[tuple[int, ...]]
        The list of qubits to idle.
    idle_side_stabilizers: list[Stabilizer]
        The stabilizers associated with the idling side.
    had_side_stabilizers: list[Stabilizer]
        The stabilizers associated with the hadamard side AFTER the transversal
        Hadamard.
    wall_stabilizers: list[Stabilizer]
        The stabilizers associated with the wall.
    idle_side_directions: DiagonalDirection
        The directions of movement of the idling side.
    had_side_directions: DiagonalDirection
        The directions of movement the hadamard side.

    Returns
    -------
    list[Stabilizer]
        The new stabilizers after the recombination of the block.
    """

    new_stabilizers = []
    stab_evolution = {}
    # Find new stabilizers for the idle side
    for stab in idle_side_stabilizers:
        new_stab = Stabilizer(
            stab.pauli,
            [qubit_mapper(dq, idle_side_directions) for dq in stab.data_qubits],
            ancilla_qubits=[
                qubit_mapper(aq, idle_side_directions) for aq in stab.ancilla_qubits
            ],
        )
        new_stabilizers.append(new_stab)
        stab_evolution[new_stab.uuid] = (stab.uuid,)
    # Find new stabilizers for the hadamard side
    for stab in had_side_stabilizers:
        new_stab = Stabilizer(
            stab.pauli,
            [qubit_mapper(dq, had_side_directions) for dq in stab.data_qubits],
            ancilla_qubits=[
                qubit_mapper(aq, had_side_directions) for aq in stab.ancilla_qubits
            ],
        )
        new_stabilizers.append(new_stab)
        stab_evolution[new_stab.uuid] = (stab.uuid,)
    # Find new stabilizers for the wall
    for stab in wall_stabilizers:
        # Some part of the wall stabilizer is on the idling side and some on the
        # hadamard side.
        # The ancilla is on the idling side
        new_stab = Stabilizer(
            stab.pauli,
            [
                (
                    qubit_mapper(q, idle_side_directions)
                    if q in qubits_to_idle
                    else qubit_mapper(q, had_side_directions)
                )
                for q in stab.data_qubits
            ],
            ancilla_qubits=[
                qubit_mapper(aq, idle_side_directions) for aq in stab.ancilla_qubits
            ],
        )
        new_stabilizers.append(new_stab)
        stab_evolution[new_stab.uuid] = (stab.uuid,)

    # Update stabilizer updates
    for new_stab_uuid, old_stabs_uuid in stab_evolution.items():
        # Only one parent per new stabilizer
        old_stab_uuid = old_stabs_uuid[0]
        if old_stab_uuid in interpretation_step.stabilizer_updates:
            interpretation_step.stabilizer_updates[new_stab_uuid] = (
                interpretation_step.stabilizer_updates.pop(old_stab_uuid)
            )

    interpretation_step.stabilizer_evolution.update(stab_evolution)
    return new_stabilizers


def get_wall_data_qubit_init(
    y_wall_initial_block: RotatedSurfaceCode,
    qubits_to_measure: list[tuple[int, ...]],
) -> dict[str, list[tuple[int, ...]]]:
    """
    Find the data qubits to initialize for the wall in the y_wall_out operation context
    and the corresponding pauli type. There is a wall qubit that is excluded from the
    initialization and the rest are initialized according to the idle side stabilizer
    that contains the qubit. Which stabilizer is chosen is determined by the qubit being
    on the right corner of a weight 4 stabilizer.

    Parameters
    ----------
    y_wall_initial_block: RotatedSurfaceCode
        The block to which the operation will be applied.
    qubits_to_measure: list[tuple[int, ...]]
        The list of qubits to measure.

    Returns
    -------
    dict[str, list[tuple[int, ...]]]
        The data qubits to initialize in the X and Z basis.
    """
    # Deduce geometric parameters
    block_orientation = y_wall_initial_block.orientation
    is_top_left_bulk_stab_x = (
        y_wall_initial_block.upper_left_4body_stabilizer.pauli_type == "X"
    )

    wall_data_qubits_to_init = {"X": [], "Z": []}
    match (block_orientation, is_top_left_bulk_stab_x):
        # pylint: disable=unnecessary-lambda-assignment
        case (Orientation.VERTICAL, False):
            # Exclude left qubit
            qubit_to_exclude = min(qubits_to_measure, key=lambda x: x[0])
            # Get the type of stab from qubit being the bottom right of some stabilizer
            lambda_to_max = lambda x: x[0] + x[1]

        case (Orientation.VERTICAL, True):
            # Exclude right qubit
            qubit_to_exclude = max(qubits_to_measure, key=lambda x: x[0])
            # Get the type of stab from qubit being the bottom left of some stabilizer
            lambda_to_max = lambda x: -x[0] + x[1]

        case (Orientation.HORIZONTAL, True):
            # Exclude bottom qubit
            qubit_to_exclude = max(qubits_to_measure, key=lambda x: x[1])
            # Get the type of stab from qubit being the top right of some stabilizer
            lambda_to_max = lambda x: x[0] - x[1]

        case (Orientation.HORIZONTAL, False):
            # Exclude top qubit
            qubit_to_exclude = min(qubits_to_measure, key=lambda x: x[1])
            # Get the type of stab from qubit being the bottom right of some stabilizer
            lambda_to_max = lambda x: +x[0] + x[1]

    for q in qubits_to_measure:
        # Skip the qubit to exclude
        if q == qubit_to_exclude:
            continue
        # Find pauli type by seeing for which stabilizer the qubit is on the appropriate
        # corner of a weight 4 stabilizer.
        # EXAMPLE:
        # For (block_orientation, is_top_left_bulk_stab_x) = (VERTICAL, False),
        # the initialization basis is the same as the pauli flavor of the stabilizer
        # whose bottom right corner is the qubit q.
        # "bottom right" is defined by the lambda function lambda_to_max.
        stab = next(
            s
            for s in y_wall_initial_block.stabilizers
            if max(s.data_qubits, key=lambda_to_max) == q and len(s.data_qubits) == 4
        )
        # Append the qubit to the corresponding list
        wall_data_qubits_to_init[stab.pauli[0]].append(q)

    return wall_data_qubits_to_init


def get_final_block_syndrome_measurement_first_swap_then_qec_cnots_circuit(
    interpretation_step: InterpretationStep,
    recombined_block: RotatedSurfaceCode,
    qubits_to_idle: list[tuple[int, ...]],
    qubits_to_had: list[tuple[int, ...]],
    idle_side_directions: DiagonalDirection,
    had_side_directions: DiagonalDirection,
    anc_qubits_to_init_idle: dict[str, list[tuple[int, ...]]],
    anc_qubits_to_init_had: dict[str, list[tuple[int, ...]]],
) -> Circuit:
    # pylint: disable=too-many-locals, too-many-arguments, too-many-positional-arguments
    # pylint: disable=too-many-branches, too-many-statements
    """
    Generate the CNOT schedule of the first SWAP-then-QEC syndrome measurement round
    of the final block of y_wall_out operation.

    Parameters
    ----------
    interpretation_step: InterpretationStep
        The interpretation step. Note that it may be mutated by generating new
        channels.
    recombined_block: RotatedSurfaceCode
        The recombined block after the y_wall_out recombination.
    qubits_to_idle: list[tuple[int, ...]]
        The list of qubits to idle.
    qubits_to_had: list[tuple[int, ...]]
        The list of qubits to Hadamard.
    idle_side_directions: DiagonalDirection
        The directions that the idling side half of the block will move.
    had_side_directions: DiagonalDirection
        The directions that the hadamard side half of the block will move.
    anc_qubits_to_init_idle: dict[str, list[tuple[int, ...]]]
        The ancilla qubits to initialize on the idling side.
    anc_qubits_to_init_had: dict[str, list[tuple[int, ...]]]
        The ancilla qubits to initialize on the hadamard side.

    Returns
    -------
    Circuit
        The CNOT circuit of the first SWAP-then-QEC syndrome measurement round
        of the final block of y_wall_out operation.
    """
    config, pivot_corners = recombined_block.config_and_pivot_corners
    if config != 3:
        raise ValueError(
            "The final block of Y-wall-out must have type-3 corner configuration."
        )

    long_end_corner, angle_corner = pivot_corners[0], pivot_corners[2]
    top_left_corner = recombined_block.upper_left_qubit
    is_horizontal = recombined_block.is_horizontal

    # Schedules
    # Z schedule is determined based on the orientation of the short edge
    region_124_z_schedule = FourBodySchedule.Z if is_horizontal else FourBodySchedule.N
    region_3_z_schedule = region_124_z_schedule.opposite_schedule()
    region_12_x_schedule = region_124_z_schedule.opposite_schedule()
    region_34_x_schedule = region_12_x_schedule.opposite_schedule()

    # Initialize the cnots
    cnots = [[] for _ in range(5)]

    # TIMESLICE 0
    # The idling part and the hadamard sides are going to meet at the wall so they need
    # to be moved in different directions

    # Find CNOTs for the hadamard side

    had_side_data_to_anc_vec = direction_to_coord(had_side_directions, 0)
    for qub in qubits_to_had:
        anc_qub = tuple(map(sum, zip(qub, had_side_data_to_anc_vec, strict=True)))
        if anc_qub in anc_qubits_to_init_had["Z"]:
            cnot_pair = (qub, anc_qub)
        elif anc_qub in anc_qubits_to_init_had["X"]:
            cnot_pair = (anc_qub, qub)
        else:
            raise ValueError("The ancilla qubit was not found.")
        cnots[0].append(cnot_pair)

    # Find CNOTs for the idling side
    idle_side_data_to_anc_vec = direction_to_coord(idle_side_directions, 0)
    for qub in qubits_to_idle:
        anc_qub = tuple(map(sum, zip(qub, idle_side_data_to_anc_vec, strict=True)))
        if anc_qub in anc_qubits_to_init_idle["Z"]:
            cnot_pair = (qub, anc_qub)
        elif anc_qub in anc_qubits_to_init_idle["X"]:
            cnot_pair = (anc_qub, qub)
        else:
            raise ValueError("The ancilla qubit was not found.")
        cnots[0].append(cnot_pair)

    # TIMESLICE 1-4
    for stab in recombined_block.stabilizers:
        # Find stabilizer's position
        pauli_type = stab.pauli_type
        (
            is_in_region_12,
            is_in_region_3,
            _,
            is_on_wall,
            is_on_hadamard_side,
            boundary_direction,
        ) = find_stabilizer_position_in_final_block(
            recombined_block, stab, pivot_corners
        )
        is_on_idle_side = not (is_on_wall or is_on_hadamard_side)

        # Find stabilizer's schedule
        if pauli_type == "Z":
            schedule = region_3_z_schedule if is_in_region_3 else region_124_z_schedule
        else:
            schedule = region_12_x_schedule if is_in_region_12 else region_34_x_schedule

        # Find the starting qubit direction
        compare_corner = angle_corner if is_on_hadamard_side else long_end_corner
        starting_diag_direction = find_relative_diagonal_direction(
            top_left_corner, compare_corner
        )
        detailed_schedule = DetailedSchedule.from_schedule_and_direction(
            schedule, starting_diag_direction
        )

        if len(stab.data_qubits) == 4:
            data_qubits = detailed_schedule.get_stabilizer_qubits(stab)
        else:
            data_qubits = detailed_schedule.get_stabilizer_qubits(
                stab, boundary_direction
            )

        for i, data_qubit in enumerate(data_qubits):
            if (not is_on_wall and i == 0) or data_qubit is None:
                continue
            anc_qubit = stab.ancilla_qubits[0]

            if pauli_type == "Z":
                # If the stabilizer is Z, cnot from data qubit to ancilla
                cnot_pair = (data_qubit, anc_qubit)
            elif pauli_type == "X":
                # If the stabilizer is X, cnot from ancilla to data qubit
                cnot_pair = (anc_qubit, data_qubit)
            else:
                raise ValueError("Unknown stabilizer type.")

            # Find correct CNOT layer
            layer = (
                i + 1
                if (
                    is_on_wall  # on wall
                    or is_on_idle_side
                    and not is_in_region_12
                    and (
                        pauli_type == "Z"
                        and i in [2, 3]
                        or pauli_type == "X"
                        and i in [1, 2, 3]
                    )  # idle side
                    or is_in_region_3
                    and (
                        pauli_type == "Z"
                        and i in [1, 2, 3]
                        or pauli_type == "X"
                        and i == 3
                    )  # region 3
                )
                else i
            )
            cnots[layer].append(cnot_pair)

    cnots_circuit = Circuit(
        name="First SWAP-then-QEC final block syndrome measurement CNOT circuit",
        circuit=[
            [
                Circuit(
                    "cx",
                    channels=[
                        interpretation_step.get_channel_MUT(q) for q in qubit_pair
                    ],
                )
                for qubit_pair in cnot_slice
            ]
            for cnot_slice in cnots
        ],
    )

    return cnots_circuit



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from __future__ import annotations

from pydantic.dataclasses import dataclass
from loom.eka.utilities import dataclass_config

from .utilities import Cbit


@dataclass(config=dataclass_config)
class LogicalObservable:
    """
    Once a logical operator is measured, the details of the measurement are stored in
    an instance of LogicalObservable. This dataclass does not store the actual value of
    the logical observable but instead the required information on how to calculate it
    given the output of a quantum circuit.

    Parameters
    ----------
    label : str
        Label of the logical observable measurement which is later on used to access the
        measurement result
    measurements : tuple[Cbit, ...]
        Tuple of classical bits from which the logical observable is calculated. The
        logical observable is calculated by taking the parity of these measurements.
        This includes the readout of data qubits as well as potential updates/
        corrections based on data qubit readouts during e.g. split or shrink operations.
    """

    label: str
    measurements: tuple[Cbit, ...]

    def __eq__(self, other: LogicalObservable) -> bool:
        return self.label == other.label and set(self.measurements) == set(
            other.measurements
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import Circuit
from loom.eka.operations import ResetAllDataQubits
from loom.eka.stabilizer import Stabilizer
from loom.interpreter.syndrome import Syndrome

from .generate_syndromes import generate_syndromes
from ..interpretation_step import InterpretationStep


def reset_all_data_qubits(
    interpretation_step: InterpretationStep,
    operation: ResetAllDataQubits,
    same_timeslice: bool,
    debug_mode: bool,  # pylint: disable=unused-argument
) -> InterpretationStep:
    """
    Resets all data qubits of a block to a specific SingleQubitPauliEigenstate.
    It also adds empty Syndrome objects for the stabilizers that will be deterministic
    in the first round of syndrome measurement cycles dependent on the
    initialization state. This helps to put Detectors on these deterministic
    measurements when the block is measured.

    NOTE: Initializing a Y state may come with some caveats, as the implementation of
    the initialization may not be fault-tolerant for some codes. For example, in the
    case of the Rotated Surface Code, initializing a Y state may require distillation
    for it to be fault-tolerant.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        Interpretation step containing the blocks whose data qubits need to be reset.
    operation : ResetAllDataQubits
        Reset data operation description.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block

    Returns
    -------
    InterpretationStep
        Interpretation step after the reset data operation.
    """

    # Get the block
    block_before_reset = interpretation_step.get_block(operation.input_block_name)
    block = block_before_reset.rename(operation.input_block_name)  # Create a new uuid

    # Create a circuit that resets the data qubits to the given state
    reset_circuit = Circuit(
        name=f"reset all data qubits of block {block.unique_label} to "
        f"|{operation.state.value}>",
        circuit=[
            # Reset the data qubits on the same time step
            [
                Circuit(
                    f"reset_{operation.state.value}",
                    channels=interpretation_step.get_channel_MUT(q),
                )
                for q in block.data_qubits
            ]
        ],
    )

    # Create single-qubit stabilizers for the reset data qubits
    reset_single_qubit_stabilizers = {
        Stabilizer(pauli=operation.state.pauli_basis, data_qubits=(q,))
        for q in block.data_qubits
    }
    # pylint: disable-next=unused-variable
    reset_single_qubit_syndromes = (
        Syndrome(
            stabilizer=stab.uuid,
            measurements=(),
            block=block.uuid,
            round=-1,  # should not be associated with any round
            labels={stab.uuid: stab.data_qubits[0]},
        )
        # only deterministic outcomes are considered
        for stab in reset_single_qubit_stabilizers
        if stab.pauli == operation.state.pauli_basis
    )

    relevant_stabs = [
        stab
        for stab in block.stabilizers
        if set(stab.pauli) == {operation.state.pauli_basis}
    ]

    initialization_values = [
        (
            (1,)
            if (
                # The parity of the first syndrome is 1 only if:
                # 1. The state is a -1 eigenstate of the corresponding pauli
                # 2. The number of qubits is odd
                operation.state.basis_expectation_value == -1
                and len(stab.pauli) % 2 == 1
            )
            else ()
        )
        for stab in relevant_stabs
    ]

    new_syndromes = generate_syndromes(
        interpretation_step=interpretation_step,
        stabilizers=relevant_stabs,
        block=block,
        stab_measurements=list(initialization_values),
    )

    interpretation_step.append_syndromes_MUT(new_syndromes)
    # Add the reset circuit to the interpretation step
    interpretation_step.append_circuit_MUT(reset_circuit, same_timeslice)

    # Change the block history
    interpretation_step.update_block_history_and_evolution_MUT(
        old_blocks=(block_before_reset,),
        new_blocks=(block,),
        update_evolution=False,
    )

    # Add `reset_single_qubit_stabilizers` for new block
    interpretation_step.update_reset_single_qubit_stabilizers_MUT(
        block_id=block.uuid, new_single_qubit_stabilizers=reset_single_qubit_stabilizers
    )

    # Return the new interpretation step
    return interpretation_step



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from loom.eka import PauliOperator
from loom.eka.utilities import Orientation
from loom.interpreter import InterpretationStep

from .y_wall_out_initial_syndrome_measurement import (
    y_wall_out_initial_syndrome_measurement,
)
from .y_wall_measurement_hadamard import (
    y_wall_out_measurement_and_hadamard,
    find_qubit_sets,
)
from .y_wall_out_recombination_swap_then_qec import (
    y_wall_out_recombination_swap_then_qec,
    get_idle_hadamard_info,
)
from .y_wall_out_final_swap_then_qec import y_wall_out_final_swap_then_qec
from .y_wall_out_final_syndrome_measurement import y_wall_out_final_qec_rounds
from ...code_factory import RotatedSurfaceCode


def y_wall_out(
    interpretation_step: InterpretationStep,
    block: RotatedSurfaceCode,
    wall_position: int,
    wall_orientation: Orientation,
    same_timeslice: bool,
    debug_mode: bool,
) -> InterpretationStep:
    """
    Implement the y_wall_out operation. The operation consists of three steps:

    - A.) Begin y_wall_out composite operation session

    - B.) Consistency check

    - C.) Extract geometric information
    
    - D.) Call the sub-operations in sequence:
        - D.1) One round of syndrome measurement of the initial block, \
        with CNOT scheduling specifically designed for fault tolerance
        - D.2) Y wall measurement and Hadamard
        - D.3) Recombination of the Block to get rid of the wall via swap-then-QEC
        - D.4) Swap-then-QEC to move the recombined block back to its original position
        - D.5) d-2 rounds of syndrome measurement of the final block, \
        with CNOT scheduling specifically designed for fault tolerance
    
    - E.) End the composite operation session and append the circuit


    Example: the block on the left is transformed into the block on the right::

                   X                                    X
           *(0,0) --- (1,0) --- (2,0)*          *(0,0) --- (1,0) --- (2,0)*
              |         |         |                |         |         |
              |    Z    |    X    |  Z             |    Z    |    X    |  Z
              |         |         |                |         |         |
            (0,1) --- (1,1) --- (2,1)            (0,1) --- (1,1) --- (2,1)
              |         |         |                |         |         |
           Z  |    X    |    Z    |             Z  |    X    |    Z    |
              |         |         |                |         |         |
            (0,2) --- (1,2) --- (2,2)     ->     (0,2) --- (1,2) --- (2,2)*
              |         |         |                |         |         |
              |    Z    |    X    |  Z             |    Z    |    X    |
              |         |         |                |         |         |
           *(0,3) --- (1,3) --- (2,3)            (0,3) --- (1,3) --- (2,3)
              |         |         |                |         |         |
              |    X    |    Z    |             Z  |    X    |    Z    |  X
              |         |         |                |         |         |
            (0,4) --- (1,4) --- (2,4)            (0,4) --- (1,4) --- (2,4)*
              |         |         |                     Z
           X  |    Z    |    X    |  Z
              |         |         |
            (0,5) --- (1,5) --- (2,5)*
                   X

    Other allowed blocks are reflection along a vertical axis and rotation by 90
    degrees. More examples in tests.

    Parameters
    ----------
    interpretation_step : InterpretationStep
        The interpretation step containing the blocks involved in the operation.
    block : RotatedSurfaceCode
        The block to which the operation will be applied.
    wall_position : int
        The position of the wall.
    wall_orientation : Orientation
        The orientation of the wall.
    same_timeslice : bool
        Flag indicating whether the operation is part of the same timestep as the
        previous operation.
    debug_mode : bool
        Flag indicating whether the interpretation should be done in debug mode.
        Activating debug mode will enable commutation validation for Block.

    Returns
    -------
    InterpretationStep
        The interpretation step after applying the y_wall_out operation.
    """

    # A) Begin y_wall_out composite operation session
    interpretation_step.begin_composite_operation_session_MUT(
        same_timeslice=same_timeslice,
        circuit_name=(f"y_wall_out operation on block {block.unique_label}"),
    )

    # B) Consistency check
    y_wall_out_consistency_check(block, wall_position, wall_orientation)

    # C) Extract geometric information
    is_wall_hor = wall_orientation == Orientation.HORIZONTAL
    qubits_to_measure, qubits_to_idle, qubits_to_hadamard = find_qubit_sets(
        block, wall_position, is_wall_hor
    )

    # D) Call the sub-operations in sequence
    # D.1) One round of syndrome measurement of the initial block
    interpretation_step = y_wall_out_initial_syndrome_measurement(
        interpretation_step,
        block,
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # D.2) Y wall measurement and Hadamard
    interpretation_step = y_wall_out_measurement_and_hadamard(
        interpretation_step,
        block,
        wall_position,
        wall_orientation,
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block: RotatedSurfaceCode = interpretation_step.get_block(
        block.unique_label
    )
    # Extract idle side directions for later use
    _, _, _, idle_side_directions, _ = get_idle_hadamard_info(
        current_block,
        qubits_to_hadamard,
        qubits_to_idle,
    )

    # D.3) Recombination of the Block to get rid of the wall via swap-then-QEC
    interpretation_step = y_wall_out_recombination_swap_then_qec(
        interpretation_step,
        current_block,
        block,
        qubits_to_idle,
        qubits_to_measure,
        qubits_to_hadamard,
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(block.unique_label)

    # D.4) Swap-then-QEC to move the recombined block back to its original position
    interpretation_step = y_wall_out_final_swap_then_qec(
        interpretation_step,
        current_block,
        idle_side_directions,
        same_timeslice=False,
        debug_mode=debug_mode,
    )
    current_block = interpretation_step.get_block(block.unique_label)

    # D.5) d-2 rounds of syndrome measurement of the final block
    interpretation_step = y_wall_out_final_qec_rounds(
        interpretation_step,
        current_block,
        same_timeslice=False,
        debug_mode=debug_mode,
    )

    # E) End the composite operation session and append the circuit
    y_wall_out_circuit = interpretation_step.end_composite_operation_session_MUT()
    interpretation_step.append_circuit_MUT(y_wall_out_circuit, same_timeslice)

    return interpretation_step


def y_wall_out_consistency_check(
    block: RotatedSurfaceCode,
    wall_pos: int,
    wall_orientation: Orientation,
) -> None:
    """
    Check if the y_wall_out operation can be applied to a given block. Note that there
    are 4 different cases defined by:
    - the orientation of the block
    - whether the top-left bulk stabilizer is X or Z

    Parameters
    ----------
    block: RotatedSurfaceCode
        The block to which the operation will be applied.
    wall_pos: int
        The position of the wall.
    wall_orientation: Orientation
        The orientation of the wall.

    Raises
    ------
    ValueError
        If the block dimensions are not valid.
        If the wall position is not valid.
        If the block and the wall do not have perpendicular orientations.
        If the left boundary of a horizontal block is not a Z-type boundary.
        If the top boundary of a vertical block is not a Z-type boundary.
        If the block does not have 3 topological corners located at the geometric
        corners.
        If the missing geometric corner is not the expected one.
        If the non-geometric topological corner is not the expected one.
        If the Z logical operator is not located at the expected position.
        If the X logical operator is not located at the expected position.
    """

    # Extract some information about the geometry of the block and the wall
    is_wall_hor = wall_orientation == Orientation.HORIZONTAL
    topological_corners = block.topological_corners
    geometric_corners = block.geometric_corners
    u_l_qub = block.upper_left_qubit
    is_u_l_stab_x = block.upper_left_4body_stabilizer.pauli[0] == "X"
    dim_x, dim_z = block.size
    larger_dim = max(dim_x, dim_z)
    smaller_dim = min(dim_x, dim_z)

    # Check the block dimensions
    if smaller_dim % 2 == 0 or larger_dim % 2 != 0:
        raise ValueError(
            "The smaller dimension of the block must be odd and the larger dimension "
            "must be even."
        )

    # Check block orientation and wall orientation
    is_block_hor = dim_x == larger_dim
    if is_block_hor == is_wall_hor:
        raise ValueError("The block and the wall must have perpendicular orientations.")

    if is_block_hor:
        if block.boundary_type("left") != "Z":
            raise ValueError(
                "The left boundary of a horizontal block must be a Z-type boundary."
            )
    else:
        if block.boundary_type("top") != "Z":
            raise ValueError(
                "The top boundary of a vertical block must be a Z-type boundary."
            )

    # Check the wall position
    if wall_pos != smaller_dim:
        raise ValueError(
            "The wall position must be such that the block on the bottom/right side "
            "of the wall is a square."
        )

    # Check the corners
    common_corners = set(topological_corners) & set(geometric_corners)
    # There should be 3 topological corners located at the geometric corners
    if len(common_corners) != 3:
        raise ValueError(
            "The block must have 3 topological corners located at the geometric "
            "corners."
        )

    # Find which topological corner is not geometric and which geometric corner is not
    # topological
    geom_corner_not_topol = next(iter(set(geometric_corners) - common_corners))
    topol_corner_not_geom = next(iter(set(topological_corners) - common_corners))

    # Find the expected non-geometric topological corner:
    # s: start of the block (left/top), e: end of the block(right/bot), w: wall_position
    # If the block is horizontal:
    #   If the block has the top left stabilizer as X:
    #      (e, w)
    #   else:
    #      (s, w)
    # else:
    #   If the block has the top left stabilizer as X:
    #      (w, e)
    #   else:
    #      (w, s)
    exp_topol_corner_not_geom = (
        u_l_qub[0] + (wall_pos if is_block_hor else is_u_l_stab_x * (smaller_dim - 1)),
        u_l_qub[1] + (is_u_l_stab_x * (smaller_dim - 1) if is_block_hor else wall_pos),
        0,
    )

    if exp_topol_corner_not_geom != topol_corner_not_geom:
        raise ValueError(
            f"The non-geometric topological corner should have been "
            f"{exp_topol_corner_not_geom} but it is {topol_corner_not_geom}."
        )

    # Find the expected non-topological geometric corner:
    # If the top left stabilizer is X:
    #   it's the botton right corner i.e. maximize(x[0] + x[1])
    # else:
    #   If the block is horizontal:
    #      it's the top right corner i.e. maximize(x[0] - x[1])
    #   else:
    #      it's the bottom left corner i.e. maximize(-x[0] + x[1])
    exp_geom_corner_not_topol = max(
        geometric_corners,
        key=lambda x: (
            x[0] + x[1] if is_u_l_stab_x else (-x[0] + x[1]) * (-1) ** (is_block_hor)
        ),
    )

    if exp_geom_corner_not_topol != geom_corner_not_topol:
        raise ValueError(
            f"The non-topological geometric corner should have been "
            f"{exp_geom_corner_not_topol} but it is {geom_corner_not_topol}."
        )

    # Check that the logical operators are located in the correct positions
    expected_z_logical = PauliOperator(
        "Z" * smaller_dim,
        block.boundary_qubits("left") if is_block_hor else block.boundary_qubits("top"),
    )
    if block.logical_z_operators[0] != expected_z_logical:
        raise ValueError(
            "The Z logical operator is not located at the expected position. It needs "
            "to be on the left for a horizontal block and on the top for a vertical "
            "block."
        )

    # The X logical operator suffices to be straight and with the right distance
    # given the geometry of the block
    x_log_data_qubits = block.logical_x_operators[0].data_qubits
    is_x_logical_vert = all(q[0] == x_log_data_qubits[0][0] for q in x_log_data_qubits)
    is_x_logical_hor = all(q[1] == x_log_data_qubits[0][1] for q in x_log_data_qubits)
    is_x_straight = is_x_logical_vert or is_x_logical_hor
    is_x_correct_distance = len(x_log_data_qubits) == smaller_dim + 1
    if not (is_x_straight and is_x_correct_distance):
        raise ValueError(
            "The X logical operator is not located at the expected position. It needs "
            "to be straight and with the smallest distance possible given the geometry "
            "of the block."
        )



"""
Copyright 2024 Entropica Labs Pte Ltd

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

"""

from pydantic.dataclasses import dataclass

from loom.eka.utilities import Direction, dataclass_config
from loom.eka.operations import CodeOperation


@dataclass(config=dataclass_config)
class AuxCNOT(CodeOperation):
    """
    Apply a CNOT operation between two blocks using a grow-split-merge-shrink sequence.

    Parameters
    ----------
    input_blocks_name : tuple[str, str]
        Names of the control and target blocks, respectively.
    """

    input_blocks_name: tuple[str, str]


@dataclass(config=dataclass_config)
class TransversalHadamard(CodeOperation):
    """
    Apply a Transversal Hadamard on a block.

    Parameters
    ----------
    input_block_name : str
        Name of the block to apply the transversal hadamard gates.
    """

    input_block_name: str


# pylint: disable=line-too-long
@dataclass(config=dataclass_config)
class MoveBlock(CodeOperation):
    """
    Move the selected block 1 unit in the chosen direction in a fault-tolerant manner.
    The SWAP operation occurs as part of the syndrome extraction round, allowing the block
    to remain fault tolerant throughout the move.

    .. note::
        1 unit is defined as the distance between two adjacent qubits (with equal unit vectors) in the block.
        For example, moving from (0, 0, 0) to (1, 0, 0) is a 1-unit move in the x-direction (i.e., to the "right").

    The possible directions are:

    1. Direction.TOP (top) – Moving the block toward the top; all qubits are shifted by (0, -1, 0)
    2. Direction.BOTTOM (bottom) – Moving the block toward the bottom; all qubits are shifted by (0, 1, 0)
    3. Direction.LEFT (left) – Moving the block toward the left; all qubits are shifted by (-1, 0, 0)
    4. Direction.RIGHT (right) – Moving the block toward the right; all qubits are shifted by (1, 0, 0)

    Parameters
    ----------
    input_block_name : str
        Name of the block to move.
    direction: Direction
        Direction in which to move the block.
    """

    input_block_name: str
    direction: Direction


@dataclass(config=dataclass_config)
class LogicalPhaseViaYwall(CodeOperation):
    """
    Apply a logical phase gate to a RotatedSurfaceCode block by:

    1. Growing it towards the right/left or bottom/top depending on the orientation of
    the X boundary of the block. If the X boundary is horizontal, the block can be
    grown to the right or left. If the X boundary is vertical, the block can be
    grown to the top or bottom.
    2. Moving corners
    3. Measuring a wall of qubits in the Y basis
    4. Getting the block back to its original size

    Parameters
    ----------
    input_block_name : str
        Name of the block to apply the logical phase gate.
    growth_direction : Direction
        Direction in which to grow the block. This is the direction of the X boundary
        of the block. If the X boundary is horizontal, the block has to be grown to the
        right or left. If the X boundary is vertical, the block has to be grown to the
        top or bottom.
    """

    input_block_name: str
    growth_direction: Direction


@dataclass(config=dataclass_config)
class RotateBlock(CodeOperation):
    """
    Rotate a block by 90 degrees in a fault-tolerant manner. This modifies both the bulk
    and the boundaries of the block.

    E.g. RotateBlock on the following rotated surface code block::

                     x                               z
            o --- o --- o                   o --- o --- o
          z |  x  |  z  |                 x |  z  |  x  |
            o --- o --- o      Rotate       o --- o --- o
            |  z  |  x  | z      ->         |  x  |  z  | x
            o --- o --- o                   o --- o --- o
               x                               z

    Parameters
    ----------
    input_block_name : str
        Name of the block to rotate.
    grow_direction : Direction
        Direction in which to grow the block so that the rotation can be done in
        a fault-tolerant manner.
    """

    input_block_name: str
    grow_direction: Direction
